"""the p8 schema models for registration"""

from .types import Function
from .types import *
from .PlanModel import PlanModel

sample_models = [
    LanguageModelApi(id = make_uuid('gpt-4o-2024-08-06'), name = 'gpt-4o-2024-08-06', scheme='openai', completions_uri='https://api.openai.com/v1/chat/completions', token_env_key='OPENAI_API_KEY'),
    LanguageModelApi(id = make_uuid('gpt-4o-mini'), name = 'gpt-4o-mini', scheme='openai', completions_uri='https://api.openai.com/v1/chat/completions', token_env_key='OPENAI_API_KEY'),
    LanguageModelApi(id = make_uuid('cerebras-llama3.1-8b'), name = 'cerebras-llama3.1-8b', model='llama3.1-8b', scheme='openai', completions_uri='https://api.cerebras.ai/v1/chat/completions', token_env_key='CEREBRAS_API_KEY'),
    LanguageModelApi(id = make_uuid('groq-llama-3.3-70b-versatile'), name = 'groq-llama-3.3-70b-versatile', model='llama-3.3-70b-versatile', scheme='openai', completions_uri='https://api.groq.com/openai/v1/chat/completions', token_env_key='GROQ_API_KEY'),
    LanguageModelApi(id = make_uuid('claude-3-5-sonnet-20241022'), name = 'claude-3-5-sonnet-20241022', scheme='anthropic', completions_uri='https://api.anthropic.com/v1/messages', token_env_key='ANTHROPIC_API_KEY'),
    LanguageModelApi(id = make_uuid('gemini-1.5-flash'), name = 'gemini-1.5-flash', scheme='google', completions_uri='https://generativelanguage.googleapis.com/v1beta/models/gemini-1.5-flash:generateContent', token_env_key='GEMINI_API_KEY'),
    LanguageModelApi(id = make_uuid('gemini-2.0-flash'), name = 'gemini-2.0-flash', scheme='google', completions_uri='https://generativelanguage.googleapis.com/v1beta/models/gemini-2.0-flash:generateContent', token_env_key='GEMINI_API_KEY'),
    LanguageModelApi(id = make_uuid('gemini-2.0-flash-thinking-exp-01-21'), name = 'gemini-2.0-flash-thinking-exp-01-21', scheme='google', completions_uri='https://generativelanguage.googleapis.com/v1beta/models/gemini-2.0-flash-thinking-exp-01-21:generateContent', token_env_key='GEMINI_API_KEY'),
    LanguageModelApi(id = make_uuid('gemini-2.0-pro-exp-02-05'), name = 'gemini-2.0-pro-exp-02-05', scheme='google', completions_uri='https://generativelanguage.googleapis.com/v1beta/models/gemini-2.0-pro-exp-02-05:generateContent', token_env_key='GEMINI_API_KEY'),
    LanguageModelApi(id = make_uuid('deepseek-chat'), name = 'deepseek-chat', scheme='openai', completions_uri='https://api.deepseek.com/chat/completions', token_env_key='DEEPSEEK_API_KEY'),
    LanguageModelApi(id = make_uuid('grok-2-latest'), name = 'grok-2-latest', scheme='openai', completions_uri='https://api.x.ai/v1/chat/completions', token_env_key='XAI_API_KEY'),    
]

 
"""
#here is an example of how to add model - you can pull the token from your env if you are ok to save in your database for testing etc.
from percolate.services import PostgresService
from percolate.models.p8 import LanguageModelApi
from percolate.utils import make_uuid
PostgresService().update_records(LanguageModelApi(id = make_uuid('grok-2-latest'), name = 'grok-2-latest', scheme='openai', completions_uri='https://api.x.ai/v1/chat/completions', token_env_key='XAI_API_KEY'))
"""


"""some simple streaming adapters
these have not been well tested for all cases but for function calling and simple text content they work.

"""
import requests
import json
import os

def stream_openai_response(r, printer=None):
    """stream the response into the expected structure but expose a printer"""
    collected_data = {
        'tool_calls': []
    }
    collected_content = ''
    observed_tool_call = False
    tool_args = {}  # {tool_id: aggregated_args}
    tool_calls = None
    current_role = None
    for line in r.iter_lines():
        if line:
            #print(line)
            # print('')
            decoded_line = line.decode("utf-8").replace("data: ", "").strip() 
            if decoded_line and decoded_line != "[DONE]":
                try:
                    json_data = json.loads(decoded_line)
                    if "choices" in json_data and json_data["choices"]:
                        #the last chunk wil not have a choice and will have usage tokens but otherwise keep the structure
                        collected_data = json_data
                        delta = json_data["choices"][0]["delta"]
                        if delta.get('role'):
                            current_role = delta['role']
                        delta['role'] = current_role
                        
                        # Check if there's content and aggregate it
                        if "content" in delta and delta["content"]:
                            new_content = delta["content"]
                            collected_content = collected_content + new_content
                            """trace it the for the bottom"""
                            delta['content'] = collected_content
                            """we aggregate the content"""
                            if printer:
                                printer(new_content)
                        else:
                            delta['content'] = collected_content
                            
                        # Check if there are tool calls and aggregate the arguments
                        if "tool_calls" in delta:
                            if tool_calls is None:
                                """first message init"""
                                tool_calls = delta['tool_calls']
                             
                            if not observed_tool_call:
                                observed_tool_call = True
                                # if printer:
                                #     printer(f'invoking {delta["tool_calls"]}')
                            for tool_call in delta["tool_calls"]:
                                if "index" in tool_call:
                                    """for each tool call, we will index into the initial and aggregate args"""
                                    tool_index = tool_call["index"]
                                    
                                    if "function" in tool_call and "arguments" in tool_call["function"]:
                                        tool_calls[tool_index]['function']['arguments'] += tool_call["function"]["arguments"]              
                
                except json.JSONDecodeError:
                    pass  # Handle incomplete JSON chunks
    
    collected_data['choices'][0]['message'] = delta
    collected_data['choices'][0]['message']['tool_calls'] = tool_calls
    collected_data['usage'] = json_data['usage']
    if printer:
        printer('\n')
    return collected_data

def stream_anthropic_response(r, printer=None):
    """stream the response into the anthropic structure we expect but expose a printer
    anthropic uses server side events
    """
    collected_data = None
    tool_args = {}  # {tool_id: {index: aggregated_args}}
    content_text = ""  # To accumulate non-tool content
    input_tokens = 0
    output_tokens = 0
    observed_tool_call = False

    event_type = None 
    content_block_type = None
    index = None
    for line in r.iter_lines():
        if line:
            decoded_line = line.decode("utf-8")
            if decoded_line[:6] == 'event:':
                event_type = decoded_line.replace("event: ", "").strip()
                continue
            else:
                decoded_line = decoded_line.replace("data: ", "").strip()

            if decoded_line and decoded_line != "[DONE]":
                try:
                    json_data = json.loads(decoded_line)
                    event_type = json_data.get("type")                    
                    # Handle message start: Initialize structure from the first message
                    if event_type == "message_start":
                        collected_data = dict(json_data['message'])
                        input_tokens = collected_data['usage']['input_tokens']
    
                    elif event_type == "content_block_start":
                        content_block_type = json_data['content_block']['type']
                        #print(content_block_type)
                        index = json_data['index']
                        if content_block_type == 'tool_use':
                            tool_content = json_data['content_block']
                            tool_content['partial_json'] = ''
                            tool_args[index]  = tool_content
                    # Handle content block deltas with text updates
                    elif event_type == "content_block_delta" and content_block_type != 'tool_use':
                        content_type = json_data["delta"].get("type")
                        if content_type == "text_delta":
                            text = json_data["delta"].get("text", "")
                            content_text += text
                            if printer:
                                printer(text)

                    # Handle tool calls and match args using the index
                    elif event_type == "content_block_delta" and content_block_type == 'tool_use':
                        tool_input = json_data["delta"].get("partial_json")
                        if tool_input:
  
                            """TODO store the aggregated json per tool and add at the end into this structure
                            example
                            {'type': 'tool_use',
                           'id': 'toolu_01GV5rqVypHCQ6Yhrfsz8qhQ',
                           'name': 'get_weather',
                           'input': {'city': 'Paris', 'date': '2024-01-16'}}],
                            """
                            if not tool_args[index]['input']:
                                tool_args[index]['input'] = ''
                            tool_args[index]['input'] += tool_input
                            
                    # Handle message delta and stop reason at the end
                    elif event_type == "message_delta":
                        output_tokens = json_data.get("usage", {}).get("output_tokens", 0)
                        collected_data['stop_reason'] = json_data.get('stop_reason')
                        collected_data['stop_sequence'] = json_data.get('stop_sequence')

                except json.JSONDecodeError:
                    pass  # Handle incomplete JSON chunks

    # Aggregate content and tool calls into the final structure
    collected_data['content'] = [{"type": "text", "text": content_text}, 
                                *list(tool_args.values())]
    # Update usage tokens
    collected_data['usage']['input_tokens'] = input_tokens
    collected_data['usage']['output_tokens'] = output_tokens
    
    return collected_data


def stream_google_response(r, printer=None):
    """takes a response from a server side event
    the gemini url for streaming should contain :streamGenerateContent?alt=sse&key=
    Server side events are added and we use a different endpoint
    for example
    https://generativelanguage.googleapis.com/v1beta/models/gemini-1.5-flash:streamGenerateContent?alt=sse&key={os.environ.get('GEMINI_API_KEY')}
    """
    current_text_parts = []
    for line in r.iter_lines():
        if line:
            # Each chunk of data is prefixed with 'data: ', so we strip that part which is the SSE header
            line = line.decode("utf-8").strip()
            
            if line.startswith("data: "):
        
                # Remove 'data: ' and parse the JSON
                json_data = json.loads(line[len("data: "):])

                # Extract the text parts from the streamed chunk and also function call args
                candidates = json_data.get("candidates", [])
                for candidate in candidates:
                    parts = candidate.get("content", {}).get("parts", [])
                    for part in parts:
                        text = part.get("text", "")
                        current_text_parts.append(text)
                        if printer:
                            printer(text)

                finish_reason = candidate.get("finishReason")
                if finish_reason == "STOP":
                    break


    return json_data


"""some direct calls"""
def request_openai(messages,functions):
    """

    """
    #mm = [_OpenAIMessage.from_message(d) for d in pg.execute(f"""  select * from p8.get_canonical_messages(NULL, '2bc7f694-dd85-11ef-9aff-7606330c2360') """)[0]['messages']]
    #request_openai(mm)
    
    """place all system prompts at the start"""
    
    messages = [m if isinstance(m,dict) else m.model_dump() for m in messages]
    url = "https://api.openai.com/v1/chat/completions"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {os.environ.get('OPENAI_API_KEY')}"
    }

    data = {
        "model": "gpt-4o-mini",
        "messages": messages,
        "tools": functions
    }
    
    return requests.post(url, headers=headers, data=json.dumps(data))
 
 
def request_anthropic(messages, functions):
    url = "https://api.anthropic.com/v1/messages"
    headers = {
        "Content-Type": "application/json",
        "x-api-key":  os.environ.get('ANTHROPIC_API_KEY'),
        "anthropic-version": "2023-06-01",
    }
    
    data = {
        "model": "claude-3-5-sonnet-20241022",
        "max_tokens": 1024,
        "messages": [m for m in messages if m['role'] !='system'],
        "tools": functions
    }
    
    system_prompt = [m for m in messages if m['role']=='system']
   
    if system_prompt:
        data['system'] = '\n'.join( item['content'][0]['text'] for item in system_prompt )
    
    return requests.post(url, headers=headers, data=json.dumps(data))

def request_google(messages, functions):
    """
    https://cloud.google.com/vertex-ai/generative-ai/docs/multimodal/function-calling
    
    expected tool call parts [{'functionCall': {'name': 'get_weather', 'args': {'date': '2024-07-27', 'city': 'Paris'}}}]
        
    #get the functions and messages in the correct scheme. the second param in get_tools_by_name takes the scheme
    goo_mm =  [d for d in pg.execute(f" select * from p8.get_google_messages('619857d3-434f-fa51-7c88-6518204974c9') ")[0]['messages']]  
    fns =  [d for d in pg.execute(f" select * from p8.get_tools_by_name(ARRAY['get_pet_findByStatus'],'google') ")[0]['get_tools_by_name']]  
    request_google(goo_mm,fns).json()
    """        
    
    system_prompt = [m for m in messages if m['role']=='system']
    

    url = f"https://generativelanguage.googleapis.com/v1beta/models/gemini-1.5-flash:generateContent?key={os.environ.get('GEMINI_API_KEY')}"
    headers = {
        "Content-Type": "application/json"
    }
    
    """important not to include system prompt - you can get some cryptic messages"""
    data = {
        "contents": [m for m in messages if m['role'] !='system']
    }
     
    if system_prompt:
        data['system_instruction'] = {'parts': {'text': '\n'.join( item['parts'][0]['text'] for item in system_prompt )}}
    
    """i have seen gemini call the tool even when it was the data if this mode is set"""
    if functions:
        data.update(
        #    { "tool_config": {
        #       "function_calling_config": {"mode": "ANY"}
        #     },
            {"tools": functions}
        )
    
    return requests.post(url, headers=headers, data=json.dumps(data))





from __future__ import annotations
from fastapi import APIRouter, FastAPI, Response, UploadFile, File, Form
from http import HTTPStatus
from fastapi.middleware.cors import CORSMiddleware
from pathlib import Path
from fastapi.responses import StreamingResponse
from starlette.responses import HTMLResponse
import json
import traceback
import typing
from pydantic import BaseModel
from .routes import set_routes
from percolate import __version__

app = FastAPI(
    title="Percolate",
    openapi_url=f"/openapi.json",
    description=(
        """Percolate server can be used to do maintenance tasks on the database and also to test the integration of APIs in general"""
    ),
    version=__version__,
    contact={
        "name": "Percolation Labs",
        "url": "https://github.com/Percolation-Labs/percolate.git",
        "email": "percolationlabs@gmail.com",
    },
    license_info={
        "name": "MIT",
        "url": "https://opensource.org/licenses/MIT",
    },
)
api_router = APIRouter()

origins = [
    "http://localhost:5008",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/", include_in_schema=False)
@app.get("/healthcheck", include_in_schema=False)
async def healthcheck():
    return {"status": "ok"}



app.include_router(api_router)
set_routes(app)


def start():
    import uvicorn

    uvicorn.run(
        f"{Path(__file__).stem}:app",
        host="0.0.0.0",
        port=5008,
        log_level="debug",
        reload=True,
    )


if __name__ == "__main__":
    """
    You can start the dev with this in the root
    uvicorn percolate.api.main:app --port 5000 --reload
    http://127.0.0.1:5000/docs
    """
    
    start()


import yaml
import json
import os
import requests
from functools import partial
from percolate.models.p8 import Function,ApiProxy
import typing
from percolate.utils import logger
from urllib.parse import urlparse
import percolate as p8


class  _ApiTokenCache:
    def __init__(self):
        self.cache = {}
        
    def get(self, key:str):
        """
        simple cached lookup of api tokens
        """    
        if key not in self.cache:
            records = p8.repository(ApiProxy).select(proxy_uri=key)
            if records:
                self.cache[key] = records[0]['token']
        return self.cache[key]
    
_ApiTokenCache = _ApiTokenCache()
    
def map_openapi_to_function(spec,short_name:str=None):
    """Map an OpenAPI endpoint spec to a function ala open AI
       you can add functions from this to the tool format with for example
       
       ```python
       fns = [map_openai_to_function(openpi_spec_json['/weather']['get'])]
       tools = [{'type': 'function', 'function': f} for f in fns]
       ```
       
       TODO: create a pydantic model for this but for now im just trying to understand it and see who complains
    """
    def _map(schema):
        """
        Recursively map the parameters containing schema to a flatter representation,
        retaining only 'type', 'description', and optional 'enum' in nested types.
        """
        if 'schema' in schema:
            schema = schema['schema']
        mapped_schema = {
            'type': schema.get('type'),
            'description': schema.get('description', '')  
        }
        
        if 'enum' in schema:
            mapped_schema['enum'] = schema['enum']

        if schema.get('type') == 'array' and 'items' in schema:
            mapped_schema['items'] = _map(schema['items'])

        if schema.get('type') == 'object' and 'properties' in schema:
            mapped_schema['properties'] = {k: _map(v) for k, v in schema['properties'].items()}

        return mapped_schema
        
    try:
        r =  {
            'name': short_name or (spec.get('operationId') or spec.get('title')),
            'description': spec.get('description') or spec.get('summary'),
            'parameters' : {
                'type': 'object',
                'properties': {p['name']:_map(p) for p in (spec.get('parameters') or [])},
                'required': [p['name'] for p in spec.get('parameters') or [] if p.get('required')]
            } 
        }
    except:
        logger.warning(f"Failing to parse {spec=}")
        raise
    return r

    
class OpenApiSpec:
    """
    The spec object parses endpoints into function descriptions
    """
    def __init__(self, uri_or_spec: str| dict, token_key:str=None, alt_host: str=None):
        """supply a spec object (dict) or a uri to one
        The alt hose can be used if you are mapping the json from one place but invoke the server on another e.g. docker or kubernetes
        """
        self._spec_uri_str = ""
        if isinstance(uri_or_spec,str):
            self._spec_uri_str = uri_or_spec
            if uri_or_spec[:4].lower() == 'http':
                uri_or_spec = requests.get(uri_or_spec)
                if uri_or_spec.status_code == 200:
                    uri_or_spec = uri_or_spec.json()
                else:
                    raise Exception(f"unable to fetch {uri_or_spec}")
            else:
                with open(uri_or_spec, "r") as file:
                    uri_or_spec = yaml.safe_load(file)
                    
        if not isinstance(uri_or_spec,dict):
            raise ValueError("Unable to map input to spec. Ensure spec is a spec object or a uri pointing to one")

        self.spec = uri_or_spec
        """going to assume HTTPS for now TODO: consider this"""
        if 'host' in self.spec:
            self.host_uri = alt_host or f"https://{self.spec['host']}"
            if 'basePath' in self.spec:
                self.host_uri += self.spec['basePath']
        else:
            """by convention we assume the uri is the path without the json file"""
            parsed_url = urlparse(self._spec_uri_str)
            self.host_uri = alt_host or  f"{parsed_url.scheme}://{parsed_url.netloc}"

        self.token_key = token_key
        """lookup"""
        self._endpoint_methods = {op_id: (endpoint,method) for op_id, endpoint, method in self}
        self.short_names = self.map_short_names()
        
    @property
    def spec_uri(self):
        return self._spec_uri_str
    
    def map_short_names(self):
        """in the context we assume a verb and endpoint is unique"""
        d = {}
        for k,v in self._endpoint_methods.items():
            endpoint, verb = v
            """we just flatten and remove leading / - it may be that we dont want to keep the trailing / that becomes _ as its just weird - but not sure rules of uniqueness """
            d[f"{verb}_{endpoint.lstrip('/').replace('/','_').replace('-','_').replace('{','').replace('}','')}"] = k
        return d
    
    def iterate_models(self,verbs: str | typing.List[str]=None, filter_ops: typing.Optional[str]=None):
        """yield the function models that can be saved to the database
        
        Args:
           verbs: a command separated list or string list of verbs e.g. get,post to filter for ingestion
           filter_ops: an operation/endpoint filter list to endpoint ids
        """
        
        """treat params"""
        verbs=verbs.split(',') if isinstance(verbs,str) else verbs
        filter_ops=verbs.split(',') if isinstance(filter_ops,str) else filter_ops
        
        ep_to_short_names = {v:k for k,v in self.short_names.items()}
        for endpoint, grp in self.spec['paths'].items():
            for method, s in grp.items():
                op_id = s.get('operationId')
                if verbs and method not in verbs:
                    continue
                if filter_ops and op_id not in filter_ops:
                    continue
                fspec = map_openapi_to_function(s,short_name=ep_to_short_names[op_id])
                yield Function(name=ep_to_short_names[op_id],
                               key=op_id,
                               proxy_uri=self.host_uri,
                               function_spec = fspec,
                               verb=method,
                               endpoint=endpoint,
                               description=s.get('description') or s.get('summary') or s.get('title'))
                    
        
    def __repr__(self):
        """
        """
        return f"OpenApiSpec({self._spec_uri_str})"
    
    def __getitem__(self,key):
        if key not in self._endpoint_methods:
            if key in self.short_names:
                key = self.short_names[key]
            else:
                raise Exception(f"{key=} could not be mapped to an operation id or shortened name verb_endpoint")
        return self._endpoint_methods[key]
    
    def get_operation_spec(self, operation_id):
        """return the spec for this function given an endpoint operation id"""
        endpoint, verb = self._endpoint_methods[operation_id]
        return self.spec['paths'][endpoint][verb]
            
    def get_endpoint_method_from_route(self, route):
        """ambiguous and uses the first"""
        op_id = {k[0]:v for v,k in self._endpoint_methods.items()}.get(route)
        return self._endpoint_methods.get(op_id)
    
    def get_endpoint_method(self, op_id):
        """pass the operation id to get the method"""
        op =  self._endpoint_methods.get(op_id)
        if not op:
            """try the reverse mapping"""
            return self.get_endpoint_method_from_route(op_id)
        return op
    
    def resolve_ref(self, ref: str):
        """Resolve a $ref to its full JSON schema."""
        parts = ref.lstrip("#/").split("/")
        resolved = self.spec
        for part in parts:
            resolved = resolved[part]
        return resolved

    def __iter__(self):
        """iterate the endpoints with operation id, method, endpoint"""
        for endpoint, grp in self.spec['paths'].items():
            for method, s in grp.items():
                op_id = s.get('operationId')
                yield op_id, endpoint, method

    def get_expanded_schema(self):
        """expand the lot map to operation id"""
        return {operation_id: self.get_expanded_schema_for_endpoint(endpoint, method)   
                for operation_id, endpoint, method in self}
            
    def get_expanded_schema_for_endpoint(self, endpoint: str, method: str):
        """Retrieve the expanded JSON schema for a given endpoint and HTTP method."""
        parameters = []
        request_body = None
        spec = self.spec
        
        method_spec = spec["paths"].get(endpoint, {}).get(method, {})

        # Process query/path/header parameters
        for param in method_spec.get("parameters", []):
            param_schema = param.get("schema", {})
            if "$ref" in param_schema:
                param_schema = self.resolve_ref(param_schema["$ref"])
            parameters.append({
                "name": param["name"],
                "in": param["in"],
                "description": param.get("description", ""),
                "schema": param_schema
            })

        # Process requestBody (e.g., for POST requests)
        if "requestBody" in method_spec:
            content = method_spec["requestBody"].get("content", {})
            if "application/json" in content:
                schema = content["application/json"].get("schema", {})
                if "$ref" in schema:
                    schema = self.resolve_ref(schema["$ref"])
                request_body = schema

        return {"parameters": parameters, "request_body": request_body}
    
    
class OpenApiService:
    def __init__(self, uri, token_or_key:str=None, spec: OpenApiSpec=None):
        """a wrapper to invoke functions"""
        
        self.uri = uri
        self.spec = spec
        """assume token but maybe support mapping from env"""
        self.token = token_or_key or _ApiTokenCache.get(uri)
         
    def invoke(self, function:Function, data:dict=None, p8_return_raw_response:bool=False, p8_full_detail_on_error: bool = False, **kwargs):
        """we can invoke a function which has the endpoint information
        
        Args:
            function: This is a wrapped model for an endpoint stored in the database
            data: this is post-like data that can be posted for testing endpoints and we can alternative between data and kwargs
            p8_return_raw_response: a debugging/testing tool to check raw
            p8_full_detail_on_error: deciding how to send output to llms WIP
            kwargs (the function call from a language model should be passed correct in context knowing the function spec
        """
        
        #endpoint, verb = self.openapi.get_endpoint_method(op_id)
        endpoint = function.endpoint
        f = getattr(requests, function.verb)
        """rewrite the url with the kwargs"""
        endpoint = endpoint.format_map(kwargs)
        endpoint = f"{self.uri}/{endpoint.lstrip('/')}"

        if data is None: #callers dont necessarily know about data and may pass kwargs
            data = kwargs
        if data and not isinstance(data,str):
            """support for passing pydantic models"""
            if hasattr(data, 'model_dump'):
                data = data.model_dump()
            data = json.dumps(data)

        headers = { } #"Content-type": "application/json"
        if self.token:
            headers["Authorization"] =  f"Bearer {self.token}"
        
        """f is verified - we just need the endpoint. data is optional, kwargs are used properly"""
        response = f(
            endpoint,
            headers=headers,
            params=kwargs,
            data=data,
        )

        try:
            response.raise_for_status()
            if p8_return_raw_response:
                return response
        
            """otherwise we try to be clever"""
            t = response.headers.get('Content-Type') or "text" #text assumed
            if 'json' in t:
                return  response.json()
            if t[:5] == 'image':
                from PIL import Image
                from io import BytesIO
                return Image.open(BytesIO(response.content))
            content = response.content
            return content.decode() if isinstance(content,bytes) else content
                        
            
        except Exception as ex:
            if not p8_full_detail_on_error:
                """raise so runner can do its thing"""
                raise Exception(json.dumps(response.json())) 
            return {
                "data": response.json(),
                "type": response.headers.get("Content-Type"),
                "status": response.status_code,
                "requested_endpoint": endpoint,
                "info": self.model_dump(),
                "exception" : repr(ex)
            }



from fastapi import FastAPI

from .admin import router as admin_router
from .tasks import router as task_router
from .content import router as content_router

def set_routes(app: FastAPI):
    app.include_router(admin_router, prefix=f"/admin", tags=["admin"])
    app.include_router(task_router, prefix=f"/tasks", tags=["tasks"])
    app.include_router(content_router, prefix=f"/content", tags=["content"])


import yaml
from glob import glob
from percolate.utils.env import STUDIO_HOME
from pydantic import BaseModel,Field,model_validator
import typing
import os 
from percolate.utils import logger

class ProjectAgents(BaseModel):
    name: str
    description: str
    functions: typing.Optional[dict] = Field(default_factory=dict)
    spec: dict
    
class ProjectApis(BaseModel):
    name:str
    uri: str = Field(alias='openapi-uri')
    verbs: typing.Optional[typing.List[str]] = Field(default_factory=list)
    filter_ops:typing.Optional[typing.List[str]] = Field(default_factory=list)
    token: typing.Optional[str] = Field(None, description="add the token env in files and we will add it from env")
    token_env: typing.Optional[str] = Field(None, description="add the token env in files and we will add the token from env",exclude=True)
    alt_host: typing.Optional[str] = Field(None, description="The alternate host can use a different domain name eg. for docker")
    
    @model_validator(mode='before')
    @classmethod
    def _val(cls, values):
        if not values.get('token') and values.get('token_env'):
            values['token'] = os.environ.get(values.get('token_env'))
        return values
    
    
class ProjectModels(BaseModel):
    name: str
    completions_uri: str
    scheme: str
    token: typing.Optional[str] = Field(None, description="add the token env in files and we will add it from env")
    token_env: typing.Optional[str] = Field(None, description="add the token env in files and we will add the token from env",exclude=True)
    
    @model_validator(mode='before')
    @classmethod
    def _val(cls, values):
        if not values.get('token') and values.get('token_env'):
            values['token'] = os.environ.get(values.get('token_env'))
        return values
    
class Project(BaseModel):
    name: str
    database: typing.Optional[str] = Field('app', description="the database to add data to for the project")
    options: dict
    models: typing.Optional[ProjectModels]
    apis: typing.Optional[typing.List[ProjectApis]]
    models: typing.Optional[typing.List[ProjectModels]]
    agents: typing.Optional[typing.List[ProjectAgents]]
    

def open_project(name:str)->dict:
    """opens a project file in studio by folder name
    the project file is a percolate.yaml but we will also compile multiple files e.g. agents, apis etc in future
    TODO: create a pydantic parser for the project file
    """
    
    
    files = list(glob(f"{STUDIO_HOME.rstrip('/')}/projects/{name}/*.yaml"))
    yaml_file_mergeable = {}
    for f in files:
        with open(f, 'r') as f:
            yaml_file_mergeable.update(yaml.safe_load(f))
            
    try:
        return Project(**yaml_file_mergeable.get('project'))
    except Exception as ex:
        logger.warning(f"Unable to parse the file -{ex} - will return the contents so you can check it ")
        return yaml_file_mergeable
    
def apply_project(project:Project|str):
    """
    logic to apply a project to the database
    """
    from percolate.utils.ingestion import add
    from percolate.utils.index import index_codebase
    from percolate.utils.env import sync_model_keys
    
    if isinstance(project,str):
        project = open_project(project)
    opts = project.options or {}
    if opts.get('sync-env'):
        sync_data = sync_model_keys()
        logger.info(f"im syncing env keys if they are set for language model providers - will add to your database - {sync_data}")
    if opts.get('index-docs'):
        index_codebase()
    
    for api in project.apis:
        add.add_api(api.name, 
                    api.uri, 
                    api.token,
                    None, 
                    verbs=api.verbs, 
                    filter_ops=api.filter_ops,
                    alt_host=api.alt_host)
        
    """TODO"""
    return {'status':'ok'}


from pydantic import Field
from functools import partial
import typing

def EmbeddedField(embedding_provider='default')->Field:
    return partial(Field, json_schema_extra={'embedding_provider':embedding_provider})

DefaultEmbeddingField = EmbeddedField()

def KeyField():
    return partial(Field, json_schema_extra={'is_key':True})


from . import utils
from .MessageStack import  MessageStack
from .AbstractModel import AbstractModel

def get_p8_models():
    """convenience to load all p8 models in the library"""
    
    from percolate.models.inspection import get_classes
    return get_classes(package="percolate.models.p8")


from .p8 import * 

def bootstrap(apply:bool = False, apply_to_test_database: bool= True, root='../../../../extension/', alt_connection:str=None):
    """util to generate the sql that we use to setup percolate"""

    from percolate.models.p8 import sample_models
    from percolate.models.utils import SqlModelHelper
    from percolate.services import PostgresService
    from percolate.utils.env import TESTDB_CONNECTION_STRING
    from percolate.models.p8.native_functions import get_native_functions
    import glob
    
    pg = PostgresService(on_connect_error='ignore')
    
    if alt_connection:
        pg = PostgresService(connection_string=alt_connection)
    elif apply_to_test_database:
        print('Using test database and will create it if it does not exist')
        apply = True
        pg._create_db('test')
        pg = PostgresService(connection_string=TESTDB_CONNECTION_STRING)
    
        
    root = root.rstrip('/')
    print('********Building queries*******')
    """build a list of models we want to init with"""
    models = [ Project, Agent, ModelField, LanguageModelApi, Function, Session, AIResponse, ApiProxy, PlanModel, Settings, PercolateAgent, IndexAudit]
        
    """compile the functions into one file"""
    with open(f'{root}/sql/01_add_functions.sql', 'w') as f:
        print(f)
        for sql in glob.glob(f'{root}/sql-staging/p8_pg_functions/**/*.sql',recursive=True):
            print(sql)
            with open(sql, 'r') as sql:
                f.write(sql.read())
                f.write('\n\n---------\n\n')

    """add base tables"""            
    with open(f'{root}/sql/02_create_primary.sql', 'w') as f:
        print(f)
        for model in models:
            f.write(pg.repository(model,on_connect_error='ignore').model_registration_script(secondary=False, primary=True))

    """add the rest"""
    with open(f'{root}/sql/03_create_secondary.sql', 'w') as f:    
        print(f)
        for model in models:
            print(model)
            f.write(pg.repository(model,on_connect_error='ignore').model_registration_script(secondary=True, primary=False))
            
        script = SqlModelHelper(LanguageModelApi).get_data_load_statement(sample_models)
        f.write('\n\n-- -----------\n')
        f.write('-- sample models--\n\n')
        f.write(script)
        
        """add native functions"""
        script = SqlModelHelper(Function).get_data_load_statement(get_native_functions())
        f.write('\n\n-- -----------\n')
        f.write('-- native functions--\n\n')
        f.write(script)
        
    if apply:
        _test_apply(root=root, pg=pg)
        
def _test_apply(root='../../../../extension/', pg = None):
    """
    these are utility test methods - but we will add them to an automated deployment test script later
    passing the database in e.g. in test mode - we will clean this up later
    """
    
    from percolate.services import PostgresService
    pg = pg or PostgresService()

    print('*****applying sql schema...******')
    print()
    root = root.rstrip('/')
   
    with open(f"{root}/sql/00_install.sql") as f:
        sql = f.read()
        pg.execute(sql)
        
    with open(f"{root}/sql/01_add_functions.sql") as f:
        sql = f.read()
        pg.execute(sql)

    with open(f"{root}/sql/02_create_primary.sql") as f:
        sql = f.read()
        pg.execute(sql)
    with open(f"{root}/sql/03_create_secondary.sql") as f:
        sql = f.read()
        pg.execute(sql)
        
    with open(f"{root}/sql/10_finalize.sql") as f:
        sql = f.read()
        pg.execute(sql)
        
    print('********done*******')


from pydantic import BaseModel
from pydantic.fields import FieldInfo
from ..AbstractModel import AbstractModel
import typing
import uuid
import datetime
import types
import json
class SqlModelHelper:
    def __init__(cls, model: BaseModel):
        cls.model:AbstractModel = AbstractModel.Abstracted(model)
        cls.table_name = cls.model.get_model_table_name()
        cls.field_names = SqlModelHelper.select_fields(model)
        cls.metadata = {}
        
    @property
    def model_name(cls):
        if cls.model:
            return cls.model.get_model_full_name()
        
    def __repr__(self):
        return f"SqlModelHelper({self.model_name})"
    
    @property
    def should_model_notify_index_update(self):
        """
        this is for most models that have either a name or an embedding
        """
        fields = self.model.model_fields
        """example of disable or force enable in special cases"""
        index_notify = self.model.model_config.get('index_notify')
        if index_notify is not None:
            return index_notify
        """always in this case for graph entity by convention unless disabled"""
        if 'name' in fields:
            return True
        """and always if there are embeddings"""
        for k,v in fields.items():
            if (v.json_schema_extra or{}).get('embedding_provider'):
                return True
        """otherwise skip it"""
        return False
        
    def create_script(cls,if_not_exists:bool = False):
        """

        Args:
            if_not_exists: by default we try and fail to create for schema migration - but initially set this to True to create a clean DB
            
        (WIP) generate tables for entities -> short term we do a single table with now schema management
        then we will add basic migrations and split out the embeddings + add system fields
        we also need to add the other embedding types - if we do async process we need a metadata server
        we also assume the schema exists for now

        We will want to create embedding tables separately and add a view that joins them
        This creates a transaction of three scripts that we create for every entity
        We should add the created at and updated at system fields and maybe a deleted one

        - key register trigger -> upsert into type-name -> on-conflict do nothing

        - we can check existing columns and use an alter to add new ones if the table exists

        """
       
        def is_optional(field):
            return typing.get_origin(field) is typing.Union and type(
                None
            ) in typing.get_args(field)
            
        fields = typing.get_type_hints(cls.model)
        field_descriptions = cls.model.model_fields
        mapping =  {k:SqlModelHelper.python_to_postgres_type(v, field_descriptions.get(k)) for k,v in fields.items()}

        id_field = 'id'
        if 'id' not in mapping:
            key_field = cls.model.get_model_key_field()
            assert key_field, "You must supply either an id or a property like name or key on the model or add json_schema_extra with an is_key property on one if your fields"
        
        """this is assumed for now"""
        
        table_name = cls.model.get_model_table_name()
        
        columns = []
        key_set = False
        for field_name, field_type in mapping.items():
            column_definition = f"{field_name} {field_type}"
            if field_name == id_field:
                column_definition += " PRIMARY KEY "
                key_set = True
            if field_name in ['name','key']:
                key_set = True
            elif not is_optional(fields[field_name]):
                column_definition += " NOT NULL"
            columns.append(column_definition)

        if not key_set:
            raise ValueError("The model input does not specify a key field. Add a name or key field or specify is_key on one of your fields")
        """add system fields"""
        for dcol in ['created_at', 'updated_at', 'deleted_at']:
            if dcol not in mapping.keys():
                columns.append(f"{dcol} TIMESTAMP DEFAULT CURRENT_TIMESTAMP")
        if 'user_id' not in mapping.keys():
            columns.append("userid UUID")

        if_not_exists_flag = '' if not if_not_exists else " IF NOT EXISTS "
        columns_str = ",\n    ".join(columns)
        create_table_script = f"""CREATE TABLE {if_not_exists_flag} {table_name} (
{columns_str}
);
DROP TRIGGER IF EXISTS update_updated_at_trigger ON {table_name};
CREATE   TRIGGER update_updated_at_trigger
BEFORE UPDATE ON {table_name}
FOR EACH ROW
EXECUTE FUNCTION update_updated_at_column();

        """
    
        #in some case we will attach the notify by calling this function attach_notify_trigger_to_table(schema_name TEXT, table_name TEXT)
              
        if cls.should_model_notify_index_update:
            create_table_script += f"""
SELECT attach_notify_trigger_to_table('{cls.model.get_model_namespace()}', '{cls.model.get_model_name()}');
            """
        
        return create_table_script

    def upsert_query(
        cls,
        batch_size: int,
        returning="*",  # ID, * etc.
        restricted_update_fields: str = None,
        id_field: str = 'id' # by convention
        # records: typing.List[typing.Any],
        # TODO return * or just id for performance
    ):
        """upserts on the ID conflict

        if deleted at set generate another query to set deleted dates for records not in the id list

        This will return a batch statement for some placeholder size. You can then

        ```
        connector.run_update(upsert_sql(...), batch_data)

        ```

        where batch data is some collection of items

        ```
        batch_data = [
            {"id": 1, "name": "Sample1", "description": "A sample description 1", "value": 10.5},
            {"id": 2, "name": "Sample2", "description": "A sample description 2", "value": 20.5},
            {"id": 3, "name": "Sample3", "description": "A sample description 3", "value": 30.5},
        ]
        ```
        """

        if restricted_update_fields is not None and not len(restricted_update_fields):
            raise ValueError("You provided an empty list of restricted field")

        """TODO: the return can be efficient * for example pulls back embeddings which is almost never what you want"""
        field_list = cls.field_names
        """conventionally add in order anything that is added in upsert and missing"""
        for c in restricted_update_fields or []:
            if c not in field_list:
                field_list.append(c)

        non_id_fields = [f for f in field_list if f != id_field]
        insert_columns = ", ".join(field_list)
        insert_values = ", ".join([f"%({field})s" for field in field_list])

        """restricted updated fields are powerful for updates 
           we can ignore the other columns in the inserts and added place holder values in the update
        """
        update_set = ", ".join(
            [
                f"{field} = EXCLUDED.{field}"
                for field in restricted_update_fields or non_id_fields
            ]
        )

        value_placeholders = ", ".join(
            [f"({insert_values})" for _ in range(batch_size)]
        )

        # ^old school way but for psycopg2.extras.execute_values below is good
        value_placeholders = "%s"

        """batch insert with conflict - prefix with a delete statement that sets items to deleted"""
        upsert_statement = f"""
        -- now insert
        INSERT INTO {cls.table_name} ({insert_columns})
        VALUES {value_placeholders}
        ON CONFLICT ({id_field}) DO UPDATE
        SET {update_set}
        RETURNING {returning};
        """

        return upsert_statement.strip()
    
    @classmethod
    def select_fields(cls, model):
        """select db relevant fields"""
        fields = []
        for k, v in model.model_fields.items():
            if v.exclude:
                continue
            attr = v.json_schema_extra or {}
            """we skip fields that are complex"""
            if attr.get("sql_child_relation"):
                continue
            fields.append(k)
        return fields   
    


    @property
    def table_has_embeddings(self)->bool:
        """TODO return true if one or more columns have embeddings"""
        return True
        
    def select_query(self, fields: typing.List[str] = None, **kwargs):
        """
        if kwargs exist we use to add predicates
        """
        fields = fields or ",".join(self.field_names)

        if not kwargs:
            return f"""SELECT { fields } FROM {self.table_name} """
        predicate = SqlModelHelper.construct_where_clause(**kwargs)
        return f"""SELECT { fields } FROM {self.table_name} {predicate}"""
    
    
    @classmethod
    def construct_where_clause(cls, **kwargs) -> str:
        """
        Constructs a SQL WHERE clause from keyword arguments.

        Args:
            **kwargs: Column-value pairs where:
                - Strings, dates, and other scalar types are treated as equality (col = %s).
                - Lists are treated as ANY operator (col = ANY(%s)).

        Returns:
            predicate string
        """
        where_clauses = []
        params = []

        for column, value in kwargs.items():
            if isinstance(value, list):

                where_clauses.append(f"{column} = ANY(%s)")
                params.append(value)
            else:

                where_clauses.append(f"{column} = %s")
                params.append(value)

        where_clause = " AND ".join(where_clauses)

        return f"WHERE {where_clause}" if where_clauses else ""
    
    def create_embedding_table_script(cls)->str:
        """
        Given a model, we create the corresponding embeddings table
        
        """
     
        Q = f"""CREATE TABLE  IF NOT EXISTS {cls.model.get_model_embedding_table_name()} (
    id UUID PRIMARY KEY,  -- Hash-based unique ID - we typically hash the column key and provider and column being indexed
    source_record_id UUID NOT NULL,  -- Foreign key to primary table
    column_name TEXT NOT NULL,  -- Column name for embedded content
    embedding_vector VECTOR NULL,  -- Embedding vector as an array of floats
    embedding_name VARCHAR(50),  -- ID for embedding provider
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP, -- Timestamp for tracking
    
    -- Foreign key constraint
    CONSTRAINT fk_source_table_{cls.model.get_model_full_name().replace('.','_').lower()}
        FOREIGN KEY (source_record_id) REFERENCES {cls.table_name}
        ON DELETE CASCADE
);
"""
        return Q
    
    def try_generate_migration_script(self, field_list: typing.List[dict]) ->str:
        """
        pass in fields like this 
        
        [{'field_name': 'name', 'field_type': 'character varying'},
        {'field_name': 'id', 'field_type': 'uuid'},
        {'field_name': 'entity_name', 'field_type': 'character varying'},
        {'field_name': 'field_type', 'field_type': 'character varying'},
        {'field_name': 'deleted_at', 'field_type': 'timestamp without time zone'},
        {'field_name': 'userid', 'field_type': 'uuid'}]
        
        We will add any new fields but we will not remove or modify existing fields yet
        """
        
        field_names = [f['field_name'] for f in field_list]
        fields = typing.get_type_hints(self.model)
        field_descriptions = self.model.model_fields
        new_fields = set(fields.keys()) - set(field_names)
        script = None
        if new_fields:
            script = ""
            for f in new_fields:
                ptype = SqlModelHelper.python_to_postgres_type(fields[f], field_descriptions.get(f))
                script += f"ALTER TABLE {self.table_name} ADD COLUMN {f} {ptype}; "
        return script
    
    @staticmethod
    def python_to_postgres_types(model: BaseModel):
        """map postgres from pydantic types
        - sometimes we use attributes on the fields to coerce otherwise we map defaults from python types
        - an example mapping would be a VARCHAR length which woudl otherwise default to TEXT
        """
        fields = typing.get_type_hints(model)
        field_descriptions = model.model_fields

        return {k:SqlModelHelper.python_to_postgres_type(v, field_descriptions.get(k)) for k,v in fields.items()}
          
    @staticmethod      
    def python_to_postgres_type(py_type: typing.Any, field_annotation: FieldInfo = None) -> str:
        """
        Maps Python types to PostgreSQL types.
        The field hints can be added as overrides to what we would use by default for types
        """

        if field_annotation:
            metadata = field_annotation.json_schema_extra or {}
            if metadata.get('varchar_length'):
                return f"VARCHAR({metadata.get('varchar_length')})"
            if metadata.get('sql_type'):
                return metadata.get('sql_type')
        
        type_mapping = {
            str: "TEXT",
            int: "INTEGER",
            float: "REAL",
            bool: "BOOLEAN",
            uuid.UUID: "UUID",
            dict: "JSON",
            list: "[]",
            set: "[]",
            tuple: "ARRAY",
            datetime.datetime: "TIMESTAMP",
            typing.Any: "TEXT",  
        }

        if py_type in type_mapping:
            return type_mapping[py_type]

        origin = typing.get_origin(py_type)
        if origin is typing.Union or origin is types.UnionType:
            sub_types = [SqlModelHelper.python_to_postgres_type(t) for t in py_type.__args__ if t is not type(None)] 
            if len(sub_types) == 1:
                return sub_types[0]  
            
            if 'UUID' in sub_types:
                return "UUID" #precedence

            union =  f"UNION({', '.join(sub_types)})" 
            
            if 'JSON' in union:
                return "JSON"
            if 'TEXT[]' in union:
                return 'TEXT[]'
            if 'TIMESTAMP' in union:
                return 'TIMESTAMP'
            
            raise Exception(f"Need to handle disambiguation for union types - {union}")

        if origin in {list, typing.List, tuple, typing.Tuple}:
            sub_type = py_type.__args__[0] if py_type.__args__ else typing.Any
            pg_type = SqlModelHelper.python_to_postgres_type(sub_type)
            """decide if we want json arrays or just json"""
            if pg_type == 'JSON':
                return pg_type
            return f"{pg_type}{type_mapping[list]}"

        if origin in {dict, typing.Dict}:
            return type_mapping[dict]

        if hasattr(py_type, 'model_dump'):
            return "JSON"

        raise ValueError(f"Unsupported type: {py_type}")
    
    def get_model_field_models(self):
        """wraps the field from model method"""
        from percolate.models.p8 import ModelField
        return ModelField.get_fields_from_model(self.model)
    
    def get_model_agent_record(self):
        """wraps the agent from model method"""
        from percolate.models.p8 import Agent
        return Agent.from_abstract_model(self.model)
    
    def serialize_for_db(cls, model: dict|BaseModel, index:int=-1):
            """this exists only to allow for generalized types
            abstract models can implement model_dump to have an alt serialization path
            """
            if isinstance(model, dict):
                data = model
            elif hasattr(model, "model_dump"):
                data = model.model_dump()
            else:
                data = vars(model) 
                
           
            def adapt(item):
                if isinstance(item, uuid.UUID):
                    return str(item)
                if isinstance(item, dict):
                    return json.dumps(item,default=str)
                if isinstance(item, list) and len(item) and isinstance(item[0],dict):
                    return json.dumps(item,default=str)
                return item
                 
            data = {k:adapt(v) for k,v in data.items()}
            
            return data
        
    def get_data_load_statement(self, records: typing.List[BaseModel]):
        """Generate the insert statement for the data"""
        
        if not isinstance(records,list):
            records = [records]
            
        def sql_repr(s):
            """WIP"""
            return repr(s) if s else 'NULL'
        
        if len(records):
            
            sample = self.serialize_for_db(records[0])
            cols = f",".join(sample.keys())
            values = ",\n ".join(  f"({', '.join([sql_repr(v) for _,v in self.serialize_for_db(obj).items() ])})" for obj in records)
            #TODO it may be that id is not always what we want
            conflicts =  f",".join([f"{f}=EXCLUDED.{f}" for f in [c for c in sample.keys() if c not in ['id']]])
            return f"""INSERT INTO {self.model.get_model_table_name()}({cols}) VALUES\n {values}
        ON CONFLICT (id) DO UPDATE SET {conflicts}   ;"""
        
    def get_register_entities_query(self):
        """get the registration script for entities that have names and add the initial entities for testing to the graph"""
        
        if 'name' in self.model.model_fields:
            return  f"""select * from p8.register_entities('{self.model.get_model_full_name()}');"""


#!/usr/bin/env python

from percolate.utils import logger
import sys
logger.remove()  
logger.add(sys.stderr, level="INFO")  # Set log level to info for typer since DEBUG is default

import typer
from typing import List, Optional
from percolate.utils.ingestion import add
from percolate.utils.env import sync_model_keys
import percolate as p8
from percolate.models.p8 import PercolateAgent


app = typer.Typer()

add_app = typer.Typer()
app.add_typer(add_app, name="add")

@add_app.command()
def api(
    uri: str = typer.Argument(..., help="The API URI"),   
    name: Optional[str] = typer.Option(None, help="A friendly optional API name - the uri will be as a default name"),
    token: Optional[str] = typer.Option(None, help="Authentication token for the API"),
    file: Optional[str] = typer.Option(None, help="File associated with the API"),
    verbs: Optional[List[str]] = typer.Option(None, help="HTTP verbs allowed (e.g., GET, POST)"),
    filter_ops: Optional[str] = typer.Option(None, help="Filter operations as a string expression")
):
    """Add an API configuration."""
    typer.echo(f"Adding API: {name}")
    typer.echo(f"URI: {uri}")
    add.add_api(name=name, uri=uri, token=token,file=file, verbs=verbs,filter_ops=filter_ops)

@add_app.command()
def env(
    sync: bool = typer.Option(False, "--sync", help="Sync environment variables from .env")
):
    """Add environment variables via key-value pairs or sync from .env file"""
    if sync:
        typer.echo('---------------------------------------------------------------------------')
        typer.echo(f"🔄 Syncing env vars from your environment for loaded models in percolate.")
        typer.echo('---------------------------------------------------------------------------')
        results = sync_model_keys()
        count = 0
        for key, result in results.items():
            if result:
                count += 1
            typer.echo(f"{'✅' if result else '❌'} {key}")
        if count:
            typer.echo('-----------------------------------------------------------')
            typer.echo(f'Added {count} keys - see the p8."LanguageModelApi" table.')
            typer.echo('-----------------------------------------------------------')
        else:
            typer.echo('-----------------------------------------------------------')
            typer.echo(f'did not find any suitable keys in your environment.')
            typer.echo('-----------------------------------------------------------')
                
@add_app.command()
def function(
    name: str,
    file: str,
    args: Optional[str] = typer.Option(None, help="Arguments for the function"),
    return_type: Optional[str] = typer.Option(None, help="Return type of the function")
):
    """Add a function configuration."""
    typer.echo(f"Adding Function: {name}")
    typer.echo(f"File: {file}")
    if args:
        typer.echo(f"Args: {args}")
    if return_type:
        typer.echo(f"Return Type: {return_type}")

@add_app.command()
def agent(
    name: str,
    endpoint: str,
    protocol: Optional[str] = typer.Option("http", help="Communication protocol (default: http)"),
    config_file: Optional[str] = typer.Option(None, help="Path to the agent configuration file")
):
    """Add an agent configuration."""
    typer.echo(f"Adding Agent: {name}")
    typer.echo(f"Endpoint: {endpoint}")
    typer.echo(f"Protocol: {protocol}")
    if config_file:
        typer.echo(f"Config File: {config_file}")


# Index command with no arguments
@app.command()
def index():
    """Index the codebase (no arguments)."""
    from percolate.utils.index import index_codebase
    index_codebase()

@app.command()
def init(
    name: str = typer.Argument("default", help="The name of the project to apply"),
):
    from percolate.utils.studio import apply_project
    typer.echo(f"I'll apply project [{name}] to the database")
    status = apply_project(name)
    
    


# Ask command with a default question parameter and flags for agent and model
@app.command()
def ask( 
    question: str = typer.Argument("What is the meaning of life?", help="The question to ask"),
    agent: str = typer.Option(None, help="The agent to use"),
    model: str = typer.Option(None, help="The model to use")
):
    
    from percolate.utils.env import DEFAULT_MODEL
    typer.echo(f"Asking percolate...")
    """temp interface todo: - trusting the database is what we want but will practice with python
    
    example after indexing 
    python percolate/cli.py ask 'are there SQL functions in Percolate for interacting with models like Claude?'
    """
    #data  = p8.repository(PercolateAgent).execute(f"""  SELECT * FROM percolate_with_agent('{question}', '{agent or 'p8.PercolateAgent'}', '{model or DEFAULT_MODEL}') """)
    from percolate.models.p8 import PercolateAgent
    from percolate.services.llm import CallingContext
    
    def printer(text):
        """streaming output"""
        print(text, end="", flush=True)  
        if text == None:
            print('')
            
    
    c = CallingContext(streaming_callback=printer)
    agent = p8.Agent(PercolateAgent)
    data = agent(question,context=c)
    typer.echo('')        
    if data:
        pass
        #typer.echo(f"Session({data[0]['session_id_out']}): {data[0]['message_response']}")
        #typer.echo(data)
    else:
        typer.echo(f"Did not get a response")

if __name__ == "__main__":
    app()



import json
import hashlib
import uuid
from . import names
from loguru import logger
from pathlib import Path
import os
from .env import get_repo_root

def uuid_str_from_dict(d):
    """
    generate a uuid string from a seed that is a sorted dict
    """
    m = hashlib.md5()
    m.update(json.dumps(d, sort_keys=True).encode("utf-8"))
    return str(uuid.UUID(m.hexdigest()))


def make_uuid(input_object: str | dict):
    """
    make a uuid from input
    """

    if isinstance(input_object, dict):
        return uuid_str_from_dict(input_object)

    return str(uuid.uuid5(uuid.NAMESPACE_DNS   , input_object))


def batch_collection(collection, batch_size):
    """Yield successive batches of size batch_size from collection. can also be used to chunk string of chars"""
    for i in range(0, len(collection), batch_size):
        yield collection[i : i + batch_size]
        

    
def split_string_into_chunks(string, chunk_size=20000):
    """simple chunker"""
    return [string[i : i + chunk_size] for i in range(0, len(string), chunk_size)]



from fastapi import APIRouter, HTTPException, BackgroundTasks
from percolate.api.auth import get_current_token
from pydantic import BaseModel, Field
from percolate.services import PostgresService
import typing
import uuid

router = APIRouter()



from fastapi import APIRouter, HTTPException
from percolate.models.p8 import Task
from percolate.api.auth import get_current_token
import uuid
from fastapi import   Depends, File, UploadFile
from percolate.services import MinioService
router = APIRouter()


@router.post("/upload/")
async def upload_file(folder:str, file: UploadFile = File(...)):
    """uploads a file to a folder"""
    try:
        # Read file and upload to MinIO
        content = await file.read()
        MinioService().add_file(f"{folder}/{file.filename}",content, file.content_type)
    
        return {"filename": f"{folder}/{file.filename}", "message": "Uploaded successfully"}
    except Exception as e:
        return {"error": str(e)}


"""keep track of the schema used to return database objects
this will evolve and change plenty in the beginning
"""
from pydantic import BaseModel, Field, model_validator
import typing
import json
import enum
"""1 responses from agent calls"""

class AskStatus(enum.Enum):
    QUESTION: str = "QUESTION"
    ERROR: str = "ERROR"
    TOOL_CALL: str = "TOOL_CALL"
    TOOL_ERROR: str = "TOOL_ERROR" #when we call a tool but it breaks
    TOOL_CALL_RESPONSE: str = "TOOL_CALL_RESPONSE"
    RESPONSE: str = "RESPONSE"
    STREAMING_RESPONSE: str = "STREAMING_RESPONSE"
    COMPLETED: str = "COMPLETED"
    
class _ToolCallFunction(BaseModel):    
    name: str
    arguments: dict | str
    
    @model_validator(mode='before')
    @classmethod
    def _try_parse(cls,values):
        try:
            values['arguments'] = json.loads(values['arguments'])
        except:
            raise
        return values
class _ToolCall(BaseModel):
    """a canonical tool call ala openai"""
    id: str = Field(description="the unique id")
    type: str = Field('function', description="the type - usually function")
    function: _ToolCallFunction = Field(description="The function name and args")
class AskResponse(BaseModel):
    """When we ask an LLM anything we try to implement a "turn" interface as below"""
    
    message_response: typing.Optional[str] = Field(None, description="A textual response from the language model ")
    tool_calls: typing.Optional[typing.List[_ToolCall]] = Field(description="(JSONB) - The tool call payload from the language model possible in a canonical format e.g. OpenAI scheme")
    tool_call_result: typing.List[dict] = Field(description="(JSONB) data result from tool calls")
    status: AskStatus = Field(description="A turn can be in one of these states")
    session_id: str = Field(description="In Percolate a session is stored in the database against a user and question. Each response is pinned to a session", alias='session_id_out')
    




"""
all the table models are added here for p8 schema and can be iterated and used to generate install scripts
we add system fields in the sql model such as timestamps and user ids
"""

import uuid
from pydantic import Field, model_validator
from pydantic.fields import FieldInfo
import typing
from ..AbstractModel import AbstractModel, BaseModel, AbstractEntityModel
from .. import inspection
from percolate.utils import make_uuid
import datetime
from .. import DefaultEmbeddingField
from percolate.utils.names import EmbeddingProviders
import percolate as p8

class Function(AbstractEntityModel):
    """Functions are external tools that agents can use. See field comments for context.
    Functions can be searched and used as LLM tools. 
    The function spec is derived from OpenAPI but adapted to the conventional used in LLMs
    """
         
    id: uuid.UUID | str = Field(description="A unique id in this case generated by the proxy and function name") 
    key: typing.Optional[str] = Field(description='optional key e.g operation id')
    name: str = Field(description="A friendly name that is unique within the proxy scope e.g. a single api or python library")
    verb:  typing.Optional[str] = Field(None,description= "The verb e.g. get, post etc")
    endpoint: typing.Optional[str] = Field(None, description="A callable endpoint in the case of REST")
    description: str = DefaultEmbeddingField('', description="A detailed description of the function - may be more comprehensive than the one within the function spec - this is semantically searchable")
    function_spec: dict = Field(description="A function description that is OpenAI and based on the OpenAPI spec")
    proxy_uri: str = Field(description='a reference to an api or library namespace that qualifies the named function')
    
    def __call__(self, *args, **kwargs):
        """
        Convenience to call any function via its proxy
        """
        return p8.get_proxy(self.proxy_uri).invoke(self, **kwargs)
        
    
    @classmethod
    def from_callable(cls, fn:typing.Callable, remove_untyped:bool=True, proxy_uri:str=None):
        def process_properties(properties: dict):
             
            untyped = []
            """google at least wont like some of these but they are redundant anyway"""
            for key, details in properties.items():
                for remove_field in ['title', 'default']:
                    if remove_field in details:
                        details.pop(remove_field)
       
                if 'anyOf' in details:
                    new_list = [t for t in details['anyOf'] if t['type']!= 'null']
                   
                    #temp ignore anything after the first type
                    if len(new_list) >= 1:
                        details['type'] = new_list[0]['type']
                        details.pop('anyOf')
                    else:
                        details.pop('anyOf')
                        details['oneOf'] = new_list
               
                if 'properties' in details:
                    process_properties(details['properties'])
                
                """for language models there is no point sending untyped information
                TODO: handle this better e.g. typed kwargs - specifying additional properties true might be enough but its unclear how to handle them anyway
                """
                if not details and remove_untyped:
                    untyped.append(key)
                

            
        def _map(f):
            """make sure the structure from pydantic is the same as used elsewhere for functions"""
            p = dict(f)
            if 'properties' in p:
                process_properties(p['properties'])
    
            name = p.pop('title')
            desc = p.pop('description') if 'description' in p else 'NO DESC'
            return {
                'name': name,
                'parameters' :p,
                'description': desc
            }
        
        """we can pass this in e.g. for native. Lib is used for lib runtime or APIs for REST"""
        if not proxy_uri:
            proxy_uri = fn.__self__ if hasattr(fn, '__self__') else 'lib'
            if not isinstance(proxy_uri,str):
                proxy_uri = inspection.get_object_id(proxy_uri)
            
        s = _map(AbstractModel.create_model_from_function(fn).model_json_schema())
        key = s['name'] if not proxy_uri else f"{proxy_uri}.{s['name']}"
        id_md5_uuid =  uuid.uuid3(uuid.NAMESPACE_DNS, key)   
        return cls(id=str(id_md5_uuid), 
                   name = s['name'],
                   key  = key, 
                   endpoint=s['name'],
                   verb='get',
                   proxy_uri=proxy_uri,
                   spec=s, 
                   function_spec=s,
                   description=s['description'],
                   #this is ignored by Function but the runtime function can carry it
                   fn=fn)
        
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({ 'key': values['name'], 'proxy_uri':values['proxy_uri']})
        return values
        
   
class ApiProxy(AbstractEntityModel):
    """A list of proxies or APIs that have attached functions or endpoints - links to proxy_uri on the Function""" 
    id: typing.Optional[uuid.UUID | str] = Field(None, description="Will default to a hash of the uri")
    name: typing.Optional[str] = Field(None, description="A unique api friendly name")
    proxy_uri: str = Field(description='a reference to an api or library namespace that qualifies the named function')
    token: typing.Optional[str] = Field(None, description="the token to save")
    
            
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('name'):
            values['name'] = values['proxy_uri']
        if not values.get('id'):
            values['id'] = make_uuid(values['proxy_uri'])
        return values
        
    
    
class LanguageModelApi(AbstractEntityModel):
    """The Language model REST Apis are stored with tokens and scheme information.
    """
    
    id: uuid.UUID | str
    name: str = Field(description="A unique name for the model api e.g cerebras-llama3.1-8b")
    model: typing.Optional[str] = Field(None, description="The model name defaults to the name as they are often the same. the name can be unique based on a provider qualfier")
    scheme: typing.Optional[str] = Field('openai', description="In practice most LLM APIs use an openai scheme - currently `anthropic` and `google` can differ")
    completions_uri: str = Field(description="The api used for completions in chat and function calling. There may be other uris for other contexts")
    token_env_key: typing.Optional[str] = Field(None, description="Conventions are used to resolve keys from env or other services. Provide an alternative key")
    token: typing.Optional[str] = Field(None, description="It is not recommended to add tokens directly to the data but for convenience you might want to")
    
        
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('model'):
            values['model'] = values['name']
        return values
        
    
class Agent(AbstractEntityModel):
    """The agent model is a meta data object to persist agent metadata for search etc"""

    id: uuid.UUID| str  
    name: str
    category: typing.Optional[str] = Field(None,description="Simple property to filter agents by categories")
    description: str = DefaultEmbeddingField(description="The system prompt as markdown")
    spec: dict = Field(description="The model json schema")
    functions: typing.Optional[dict] = Field(description="The function that agent can call",default_factory=dict)
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        """we take these from the class and save them"""
        if not values.get('functions') and hasattr(cls, 'get_model_functions'):
            values['functions'] = cls.get_model_functions()
        return values
    
    
    def from_abstract_model(cls: BaseModel):
        """Given any pydantic model that behaves like an AbstractModel, get the meta object i.e. Agent"""
        cls:AbstractModel = AbstractModel.Abstracted(cls)
        name = cls.get_model_full_name()
        return Agent(id=make_uuid(name), 
                     name=name, 
                     functions=cls.get_model_functions(),
                     spec = cls.model_json_schema(),
                     description=cls.get_model_description())

class ModelField(AbstractEntityModel):
    """Fields are each field in any saved model/agent. 
    Fields are useful for describing system info such as for embeddings or for promoting.
    """
            
    id: typing.Optional[uuid.UUID|str] = Field(description="a unique key for the field e.g. field and entity key hashed")
    name: str = Field(description="The field name")
    entity_name: str
    field_type: str
    embedding_provider: typing.Optional[str] = Field(None, description="The embedding could be a multiple in future")
    description: typing.Optional[str] = None
    is_key: typing.Optional[bool] = Field(default=False, description="Indicate that the field is the primary key - our convention is the id field should be the primary key and be uuid and we use this to join embeddings")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        """this is a convenience for 'dont care' cases where you want an embedding but dont know the model name"""
        if values.get('embedding_provider') == 'default':
            values['embedding_provider'] = EmbeddingProviders.embedding_text_embedding_ada_002
        if not values.get('id'):
            values['id'] = make_uuid({'name': values['name'], 'entity_name': values['entity_name']})
        return values
    
    @classmethod
    def from_field_info(cls, key:str, field_type: FieldInfo, parent_model:BaseModel):
        """five a FieldInfo object map the field"""
        
        parent_model:AbstractModel = AbstractModel.Abstracted(parent_model)
        json_args = field_type.json_schema_extra or {}
        arg = inspection.get_innermost_args(field_type.annotation)
        arg = str(arg.__name__) if hasattr(arg, '__name__') else str(arg)
        return ModelField(name=key, 
                           entity_name= parent_model.get_model_full_name(), 
                           field_type= arg,
                           description=field_type.description,
                           **json_args)
        
    @staticmethod
    def get_fields_from_model(cls:BaseModel):
        """for any pydantic model generate the field collection"""
        cls = AbstractModel.Abstracted(cls)
        return [ModelField.from_field_info(key, value,parent_model=cls) for key, value in cls.model_fields.items()]
    
 

class TokenUsage(AbstractModel):
    """Tracks token usage for language model interactions"""
    id: uuid.UUID| str  
    model_name: str
    tokens: typing.Optional[int] = Field(0,description="the number of tokens consumed in total")
    tokens_in: typing.Optional[int] = Field(0, description="the number of tokens consumed for input")
    tokens_out: typing.Optional[int] = Field(0, description="the number of tokens consumed for output")
    tokens_other: typing.Optional[int] = Field(0, description="the number of tokens consumed for functions and other metadata")
    session_id: typing.Optional[uuid.UUID| str  ] = Field(None, description="Session id for a conversation")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('tokens'):
            values['tokens'] = (values.get('tokens_in') or 0) + (values.get('tokens_out') or 0)
        return values
    
class IndexAudit(TokenUsage):
    model_config = {
        'index_notify': False        
    }
    """Track requests to build smart indexes such as graph links or text and image embeddings"""
    metrics: typing.Optional[dict] = Field(description="metrics for records indexed",default_factory=dict)
    status: str = Field(description="Status code such as OK|ERROR")
    message: typing.Optional[str] = Field(description="Any message such as an error")
    entity_full_name: str
    
    
class _OpenAIMessage(BaseModel):
    role: str
    content: typing.Optional[str] = Field('', description='text content')
    tool_calls: typing.Optional[typing.List[dict]|dict]
    tool_call_id: typing.Optional[str] = None
    
    @classmethod
    def from_message(cls, values):
        """
        
        """
        return _OpenAIMessage(**values, tool_call_id=(values.get('tool_eval_data') or {}).get('id'))
    
    @model_validator(mode='before')
    @classmethod
    def _val(cls,values):
        if tool_calls:= values.get('tool_calls'):
            if isinstance(tool_calls,dict):
                values['tool_calls'] = [tool_calls]
        return values

class AIResponse(TokenUsage):
    """Each atom in an exchange between users, agents, assistants and so on. 
    We generate questions with sessions and then that triggers an exchange. 
    Normally the Dialogue is round trip transaction.
    """
    id: uuid.UUID| str  
    role: str = Field(description="The role of the user/agent in the conversation")
    content: str = DefaultEmbeddingField(description="The content for this part of the conversation") #TODO we may not want to automatically generate embeddings for this table
    status: typing.Optional[str] = Field(description="The status of the session such as REQUEST|RESPONSE|ERROR|TOOL_CALL|STREAM_RESPONSE")
    tool_calls: typing.Optional[typing.List[dict]|dict] = Field(default=None, description="Tool calls are requests from language models to call tools")
    tool_eval_data: typing.Optional[dict] = Field(default=None, description="The payload may store the eval from the tool especially if it is small data")
    verbatim: typing.Optional[dict|typing.List[dict]] = Field(default=None, description="the verbatim message from the language model - we dont serialized this", exclude=True)     
    function_stack: typing.Optional[typing.List[str]] = Field(None, description='At each stage certain functions are available to the model - useful to see what it has and what it chooses and to reload stack later')
 
    def to_open_ai_message(self):
        """the message structure for the scheme
        the one thing we need to do is make sure that tool calls reference the id(s)
        """
        return _OpenAIMessage.from_message(self.model_dump())
    
    @classmethod
    def from_open_ai_response(cls, response, sid:str)  :
        """"""
        choice = response['choices'][0]
        usage = response['usage']
        message = choice['message']
        tool_calls= message.get('tool_calls')
        return cls(
            id= str(uuid.uuid1()),
            role = message['role'],
            tool_calls = tool_calls,
            content=message.get('content') or '',
            tokens_in=usage['prompt_tokens'],
            tokens_out=usage['completion_tokens'],
            model_name = response['model'],
            status='TOLL_CALL' if not tool_calls else 'RESPONSE',
            session_id=sid
        )
class Session(AbstractModel):
    """Tracks groups if session dialogue"""
    id: uuid.UUID| str  
    query: typing.Optional[str] = Field(None,description='the question or context that triggered the session')
    user_rating: typing.Optional[float] = Field(None, description="We can in future rate sessions to learn what works")
    agent: str = Field("Percolate always expects an agent but we support passing a system prompt which we treat as anonymous agent")
    parent_session_id:typing.Optional[uuid.UUID| str] = Field("A session is a thread from a question+prompt to completion. We maw span child sessions")
    
    thread_id: typing.Optional[str] = Field(None,description='An id for a thread which can contain a number of sessions')
    channel_id: typing.Optional[str] = Field(None,description='The channel through which the user came. It could be a messaging platform channel or a system division')
    channel_type: typing.Optional[str] = Field(None,description='The channel type')
    metadata: typing.Optional[dict] = Field(default_factory=dict, description="Arbitrary metadata")
    session_completed_at: typing.Optional[datetime.datetime] = Field(default=None,description="An audit timestamp to write back the completion of the session - its optional as it should manage the last timestamp on the AI Response's for the session")
    
class SessionEvaluation(AbstractModel):
    """Tracks groups if session dialogue"""
    id: uuid.UUID| str  
    rating: float = Field(None,description="A rating from 0 to 1 - binary thumb-up/thumbs-down are 0 or 1")
    comments: typing.Optional[str] = Field(None, description="Additional feedback comments from the user")
    
class ModelMatrix(AbstractModel):
    """keep useful json blobs for model info"""
    
    id: uuid.UUID| str  
    entity_name: str = Field(description="The name of the entity e.g. a model in the types or a user defined model")
    enums: typing.Optional[dict] = Field(None, description="The enums used in the model - usually a job will build these and they can be used in query prompts")
    example_sql_queries: typing.Optional[dict] = Field(None, description="A mapping of interesting question to SQL queries - usually a job will build these and they can be used in query prompts")

class Project(AbstractEntityModel):
    """A project is a broadly defined goal with related resources (uses the graph)"""
    id: uuid.UUID| str  
    name: str = Field(description="The name of the entity e.g. a model in the types or a user defined model")
    description: str = DefaultEmbeddingField(description="The content for this part of the conversation")
    target_date: typing.Optional[datetime.datetime] = Field(None, description="Optional target date")
 
class Task(Project):
    """Tasks are sub projects. A project can describe a larger objective and be broken down into tasks"""
    id: typing.Optional[uuid.UUID| str] = Field(None,description= 'id generated for the name and project - these must be unique or they are overwritten')
    project_name: typing.Optional[str] = Field(None, description="The related project name of relevant")
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({'name': values['name'], 'project_name': values['project_name']})
        return values
    
    @classmethod
    def get_model_functions(cls):
        """fetch task external functions"""
        return {'get_tasks_task_name_comments': 'get comments associated with this task, supplying the task name' }
    
class User(AbstractEntityModel):
    """save user info"""
    id: uuid.UUID| str  
    name: str
    
class ContentIndex(AbstractModel):
    """Generic parsed content from files etc. added to the system"""
    id: typing.Optional[uuid.UUID| str] = Field("The id is generated as a hash of the required uri and ordinal")  
    name: typing.Optional[str] = Field(None, description="A short content name - non unique - for example a friendly label for a chunked pdf document")
    category: typing.Optional[str] = Field(None, description="A content category")
    content: str = DefaultEmbeddingField(description="The chunk of content from the source")
    ordinal: int = Field(None, description="For chunked content we can keep an ordinal")
    uri: str = Field("An external source or content ref e.g. a PDF file on blob storage or public URI")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({'uri': values['uri'], 'ordinal': values['ordinal']})
        return values
    
class PercolateAgent(ContentIndex):
    """The percolate agent is the guy that tells you about Percolate which is a multi-modal database for managing AI in the data tier.
    You can learn about the philosophy of Percolate or ask questions about the docs and codebase.
    You can lookup entities of different types or plan queries and searches.
    You can call any registered apis and functions and learn more about how they can be used.
    Call the search function to get data about Percolate
    """
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({'uri': values['uri'], 'ordinal': values['ordinal']})
        return values


class Settings(AbstractModel):
    """settings are key value pairs for percolate admin"""
    id: typing.Optional[uuid.UUID| str] = Field("The id is generated as a hash of the required key and ordinal")  
    key: str = Field(default="The key for the value to store - id is generated form this")
    value: str = Field(description="Value of the setting")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        values['id'] = make_uuid({"key": values['key']})
        return values
    


"""although percolate will push the database use case, we can use llms in the app tier"""



from pydantic import BaseModel, model_validator, Field
import typing
import json
class FunctionCall(BaseModel):
    name: str
    arguments: str | dict
    id: str
    scheme: typing.Optional[str] = Field(None, description="information about the scheme for formatting purposes")
    def get_tool_role(self):
        """the role for tool use"""
        if self.scheme == 'anthropic':
            return 'user'
        if self.scheme == 'google':
            return 'user'
        return "tool"
    @model_validator(mode='before')
    @classmethod
    def _val(cls, values):
        if isinstance(values['arguments'], str):
            values['arguments'] = json.loads(values['arguments'])
        return values
  
      
def _check_all(question  = "what is the capital of ireland and give me one fun fact too", filter_model_keys:typing.List[str]=None, functions: typing.List[dict]=None):
    """this is a simple wrapper to illustrate probing each model with a question
    Args:
        question: any question
        filter_model_keys: this is the key used in the percolate db p8.LanguageModelApi.name. This is sometimes the same as the model name but not always
    """
    from percolate import get_language_model_settings
    from percolate.models.p8 import LanguageModelApi
    
    if filter_model_keys and not isinstance(filter_model_keys,list):
        filter_model_keys = [filter_model_keys]
    models = [LanguageModelApi(**p) for p in get_language_model_settings()]
    responses = {}
    for m in models:
        if filter_model_keys and m.name not in filter_model_keys:
                continue
        llm = LanguageModel(m.name)
        responses[m.name] = llm.call_api_simple(question,functions=functions)
    return responses


from .CallingContext import CallingContext
from .LanguageModel import LanguageModel
from .MessageStackFormatter import MessageStackFormatter


"""wrap all language model apis - use REST direct to avoid deps in the library
This is a first draft - will map this to lean more on the database model 
"""

import requests
import json
import os
import typing
from .CallingContext import CallingContext
from percolate.models import MessageStack
from percolate.services import PostgresService
from percolate.models.p8 import AIResponse
import uuid
from percolate.utils import logger
import traceback
from .MessageStackFormatter import MessageStackFormatter
from .utils import *


ANTHROPIC_MAX_TOKENS_IN = 8192

class OpenAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str,  model_name:str, streaming_callback:typing.Callable=False)->AIResponse:
        """
        parse the response into our canonical format - if streaming callback send to print
        
        example tool call response
        ```
        'tool_calls': [{'id': 'call_0KPgsQaaso8IXPUpG6ktM1DC',
        'type': 'function',
        'function': {'name': 'get_weather',
            'arguments': '{"city":"Dublin","date":"2023-10-07"}'}}],
        ```
        """
        def adapt(t):
                """we want something we can call and also something to construct the message that is needed for the tool call"""
                f = t['function']
                return   {'name': f['name'], 'arguments':f['arguments'], 'id': t['id']} 
            
        try:
            if streaming_callback:
                logger.debug(f"Streaming response")
                response = stream_openai_response(response,printer=streaming_callback)
            else:
                response = response.json()
            
           
            if response.get('error'):
                logger.warning(f"Error response {response['error'].get('message')}")
                return AIResponse(id = str(uuid.uuid1()),model_name=model_name, tokens_in=0,tokens_out=0, role='assistant',
                                session_id=sid, content=response['error'].get('message'), status = 'ERROR')
            choice = response['choices'][0]
            tool_calls = choice['message'].get('tool_calls') or []
            tool_calls = [adapt(t) for t in tool_calls]
            
            return AIResponse(id = str(uuid.uuid1()),
                    model_name=response['model'],
                    tokens_in=response['usage']['prompt_tokens'],
                    tokens_out=response['usage']['completion_tokens'],
                    session_id=sid,
                    verbatim=choice['message'],
                    role=choice['message']['role'],
                    content=choice['message'].get('content') or '',
                    status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                    tool_calls=tool_calls)
        except Exception as ex:
            logger.warning(f"unexpected structure in OpenAI scheme message {response=} - caused the error {ex}")
            raise 
                    
class AnthropicAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str,  model_name:str, streaming_callback:typing.Callable=False )->AIResponse:
        """parse the response into our canonical format - if streaming callback send to print"""
        if streaming_callback:
            response = stream_anthropic_response(response,printer=streaming_callback)
        else:
            response = response.json()
        choice = response['content'] 
        def adapt(t):
            """anthropic map to our interface"""
            return {'name': t['name'], 'arguments':t['input'], 'id': t['id'], 'scheme': 'anthropic'}
        verbatim = [t for t in choice if t['type'] == 'tool_use']
        tool_calls = [adapt(t) for t in verbatim]
        if verbatim:
            """when tools are used we need a verbatim message with tool call??"""
            verbatim = {
                'role': response['role'],
                'content': response['content']
            }
        
        content = "\n".join([t['text'] for t in choice if t['type'] == 'text']) 
        return AIResponse(id = str(uuid.uuid1()),
                model_name=response['model'],
                tokens_in=response['usage']['input_tokens'],
                tokens_out=response['usage']['output_tokens'],
                session_id=sid,
                role=response['role'],
                content=content or '',
                verbatim=verbatim ,
                status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                tool_calls=tool_calls)
        
class GoogleAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str, model_name:str, streaming_callback:typing.Callable=False)->AIResponse:
        """parse the response into our canonical format - if streaming callback send to print"""
        if streaming_callback:
            response = stream_google_response(response,printer=streaming_callback)
        else:
            response= response.json()
        message = response['candidates'][0]['content']
        choice = message['parts']
        content_elements = [p['text'] for p in choice if p.get('text')]
        def adapt(t):
            return {'function': {'name': t['name'], 'arguments':t['args'], 'id': t['name'],'scheme': 'google'}}
        tool_calls = [adapt(p['functionCall']) for p in choice if p.get('functionCall')]
        tool_calls = [t['function'] for t in tool_calls]
        return AIResponse(id = str(uuid.uuid1()),
                model_name=model_name, #does not seem to return it which is fair
                tokens_in=response['usageMetadata']['promptTokenCount'],
                tokens_out=response['usageMetadata']['candidatesTokenCount'],
                session_id=sid,
                role=message['role'],
                content=',\n'.join(content_elements),
                verbatim=message if tool_calls else None,
                status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                tool_calls=tool_calls)

class LanguageModel:
    """the simplest language model wrapper we can make"""
    def __init__(self, model_name:str):
        """"""
        self.model_name = model_name
        self.db = PostgresService()
        #TODO we can use a cache for this in future
        self.params = self.db.execute('select * from p8."LanguageModelApi" where name = %s ', (model_name,))
        if not self.params:
            raise Exception(f"The model {model_name} does not exist in the Percolate settings")
        self.params = self.params[0]
        if not self.params['token']:
            """if the token is not stored in the database we use whatever token env key to try and load it from environment"""
            self.params['token'] = os.environ.get(self.params['token_env_key'])
            if not self.params['token']:
                raise Exception(f"There is no token or token key configured for model {self.model_name} - you should add an entry to Percolate for the model using the examples in p8.LanguageModelApi")
        """we use the env in favour of what is in the store"""
        self.params['token'] = os.environ.get(self.params['token_env_key']) if self.params.get('token_env_key') else self.params.get('token') 
        
    def parse(self, response: requests.models.Response, context: CallingContext=None) -> AIResponse:
        """the llm response form openai or other schemes must be parsed into a dialogue.
        this is also done inside the database and here we replicate the interface before dumping and returning to the executor
        """
        streaming_callback = context.streaming_callback if context and context.streaming_callback else None

        try:
            if response.status_code not in [200,201]:
                logger.warning(f"There was an error requesting - {response.content}")
                pass #do something for errors
 
            sid = None if not context else context.session_id
            """check the HTTP response first"""
            if self.params.get('scheme') == 'google':
                return GoogleAIResponseScheme.parse(response, sid=sid, model_name=self.model_name,streaming_callback=streaming_callback)
            if self.params.get('scheme') == 'anthropic':
                return AnthropicAIResponseScheme.parse(response,sid=sid,model_name=self.model_name,streaming_callback=streaming_callback)
            return OpenAIResponseScheme.parse(response,sid=sid, model_name=self.model_name,streaming_callback=streaming_callback)
        except Exception as ex:
            logger.warning(f"failing to parse {response=} {traceback.format_exc()}")
        
    def __call__(self, messages: MessageStack, functions: typing.List[dict], context: CallingContext=None ,debug_response:bool=False ) -> AIResponse:
        """call the language model with the message stack and functions"""
            
        response = self._call_raw(messages=messages, functions=functions,context=context)
        
        logger.debug(f"{response=}")
        if debug_response:
            return response
        
        """for consistency with DB we should audit here and also format the message the same with tool calls etc."""
        response = self.parse(response,context=context)
        return response
    
    def ask(self, question:str, functions: typing.List[dict]=None, system_prompt: str=None,context: CallingContext=None, **kwargs):
        """simple check frr question. our interface normally uses MessageStack and this is a more generic way
        
        If you want to stream create a context 
        Args:
            question: any question
            system_prompt: to test altering model output behaviour
            functions: optional list of functions in the OpenAPI like scheme
        """
        return self.__call__(MessageStack(question,system_prompt=system_prompt), functions=functions, context=context, **kwargs)
        
        
    def _call_raw(self, messages: MessageStack, functions: typing.List[dict], context: CallingContext=None):
         """the raw api call exists for testing - normally for consistency with the database we use a different interface"""
   
         return self.call_api_simple(messages.question, 
                                    functions=functions,
                                    system_prompt=messages.system_prompt, 
                                    data_content=messages.data,
                                    is_streaming=(context and context.is_streaming))

    @classmethod 
    def from_context(cls, context: CallingContext) -> "LanguageModel":
        return LanguageModel(model_name=context.model)
    
      
    def _elevate_functions_to_tools(self, functions: typing.List[dict]):
        """dialect of function wrapper for openai scheme tools"""
        return [{'type': 'function', 'function': f} for f in functions or []]
          
    def _adapt_tools_for_anthropic(self, functions: typing.List[dict]):
        """slightly different dialect of function wrapper - rename parameters to input_schema"""
        def _rewrite(d):
            return {
                'name' : d['name'],
                'input_schema': d['parameters'],
                'description': d['description']
            }
 
        return [_rewrite(d) for d in functions or []]
    
    def call_api_simple(self, 
                        question:str, 
                        functions: typing.List[dict]=None, 
                        system_prompt:str=None, 
                        data_content:typing.List[dict]=None,
                        is_streaming:bool = False,
                        temperature: float = 0.01,
                        **kwargs):
        """
        Simple REST wrapper to use with any language model
        """
        logger.debug(f"invoking model {self.model_name}")
        """select this from the database or other lookup
        e.g. db.execute('select * from "LanguageModelApi" where name = %s ', ('gpt-4o-mini',))[0]
        """
        params = self.params
        data_content = data_content or []
        
        """we may need to adapt this e.g. for the open ai scheme"""
        tools = functions or None
      
        url = params["completions_uri"]
        """use the env first"""
         
        token = params['token']
        if not token or len(token)==0:
            raise Exception(f"There is no API KEY in the env or database for model {self.model_name} - check ENV {params.get('token_env_key')}")
        headers = {
            "Content-Type": "application/json",
        }
        if params['scheme'] == 'openai':
            headers["Authorization"] = f"Bearer {token}"
            tools = self._elevate_functions_to_tools(functions)
        if params['scheme'] == 'anthropic':
            headers["x-api-key"] = token
            headers["anthropic-version"] = self.params.get('anthropic-version', "2023-06-01")
            tools = self._adapt_tools_for_anthropic(functions)
        if params['scheme'] == 'google':
            url = f"{url}?key={token}" if not is_streaming else f"{url.replace('generateContent','streamGenerateContent')}?alt=sse&key={token}"
        data = {
            "model": params['model'],
            "messages": [
                *[{'role': 'system', 'content': s} for s in [system_prompt] if s],
                {"role": "user", "content": question},
                #add in any data content into the stack for arbitrary models
                *data_content
            ],
            "tools": tools,
            'temperature': temperature
        }
        if is_streaming:
            data['stream'] = True
            data["stream_options"] = {"include_usage": True}
            
        if params['scheme'] == 'anthropic':
            data = {
                "model": params['model'],
                'temperature': temperature,
                "messages": [
                    {"role": "user", "content": question}, 
                    #because they use blocks https://docs.anthropic.com/en/docs/build-with-claude/tool-use#example-of-empty-tool-result
                    *[MessageStackFormatter.adapt_tool_response_for_anthropic(d) for d in data_content if d]
                ]
            }
            if tools:
                data['tools'] = tools
            if system_prompt:
                data['system'] = system_prompt
            if is_streaming:
                data['stream'] = True
                
            data["max_tokens"] = kwargs.get('max_tokens',ANTHROPIC_MAX_TOKENS_IN)
             
        if params['scheme'] == 'google':
            data_content = [MessageStackFormatter.adapt_tool_response_for_google(d) for d in data_content if d]
            optional_tool_config ={}#{ "tool_config": {   "function_calling_config": {"mode": "ANY"}  }  } #this seems to confuse the googs
            data = {
                "contents": [
                    {"role": "user", "parts": [{'text': question}]},
                    *data_content
                ],
                "tools": [{'function_declarations': tools}] if tools else None,
            }
            """gemini is stupid if you add a tool to use but it already has the answer
            "role": "user",  "parts": [{   "functionResponse": { <--- if the message is like this disable tools
            """
            if not data_content or 'functionResponse' not in data_content[-1]['parts'][0]:
                data.update(optional_tool_config)
            if system_prompt:
                data["system_instruction"] =  { "parts": { "text": system_prompt } }
                    
        logger.trace(f"request {data=}")
   
        response =  requests.post(url, headers=headers, data=json.dumps(data),stream=is_streaming)
        
        if response.status_code not in [200,201]:
            logger.warning(f"failed to submit: {response.status_code=}  {response.content}")
            
        return response
    
    
    



 
from fastapi import APIRouter, HTTPException
from percolate.models.p8 import Task
from percolate.api.auth import get_current_token
import uuid
from fastapi import   Depends
from pydantic import BaseModel

router = APIRouter()

class CompletionsRequest(BaseModel):
    """the OpenAPI scheme completions wrapper for Percolate"""
    model:str
    #TODO
    
@router.post("/completions")
async def get_tasks(request: CompletionsRequest, user: dict = Depends(get_current_token)):
    pass

class SimpleAskRequest(BaseModel):
    """the OpenAPI scheme completions wrapper for Percolate"""
    model:str
    question:str
    agent: str
    #TODO
    
@router.post("/")
async def ask_simple(request: SimpleAskRequest, user: dict = Depends(get_current_token)):
    pass

 


from fastapi import Depends, FastAPI, Header, HTTPException, UploadFile
from typing import Annotated, List
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from percolate.utils.env import load_db_key

bearer = HTTPBearer()
async def get_current_token(
    credentials: HTTPAuthorizationCredentials = Depends(bearer),
):
    token = credentials.credentials

    if token != load_db_key("P8_API_KEY"):
        raise HTTPException(
            status_code=401,
            detail="Invalid API KEY in token check.",
        )

    return token


 
from fastapi import APIRouter, HTTPException, Query, Path, Response
from percolate.models.p8 import Task
from percolate.api.auth import get_current_token
import uuid
from fastapi import   Depends
import typing

router = APIRouter()

@router.get("/")
async def get_tasks(user: dict = Depends(get_current_token))->typing.List[Task]:
    return Response('Dummy response')

@router.post("/", response_model=Task)
async def create_task(task: Task)->Task:
    return task

@router.get("/{task_name}/comments")
async def get_task_comments_by_name(task_name: str = Path(..., description="The unique name of the task"))->typing.List[dict]:
    """Fetch the comments related to this task if you know its entity name"""
    return [{
        'user': 'dummy_user',
        'comment': 'dummy_comment'
    },{
        'user': 'dummy_user',
        'comment': 'dummy_comment'
    }]

@router.get("/{task_name}",response_model=Task)
async def get_task_by_name(task_name: str = Path(..., description="The unique name of the task"))->Task:
    """Retrieve a task by name"""
    return {}



# @router.put("/{task_name}")
# async def update_task(task_name: str, task: Task):
#     pass

# @router.delete("/{task_id}")
# async def delete_task(draft_id: uuid.UUID):
#     pass



from pydantic import BaseModel
from percolate.models.AbstractModel import ensure_model_not_instance, AbstractModel
from percolate.models.utils import SqlModelHelper
from percolate.utils import logger, batch_collection
import typing
from percolate.models.p8 import ModelField,Agent, IndexAudit
import psycopg2
from percolate.utils.env import POSTGRES_CONNECTION_STRING , POSTGRES_DB, POSTGRES_SERVER,DEFAULT_CONNECTION_TIMEOUT
import psycopg2.extras
from psycopg2 import sql
from psycopg2.errors import DuplicateTable
from tenacity import retry, stop_after_attempt, wait_fixed
import traceback
import uuid

class PostgresService:
    """the postgres service wrapper for sinking and querying entities/models"""

    def __init__(self, model: BaseModel = None, connection_string=None, on_connect_error: str = None):
        try:
            self._connection_string = connection_string or POSTGRES_CONNECTION_STRING
            self.conn = None
            self.helper = SqlModelHelper(AbstractModel)  
            if model:
                """we do this because its easy for user to assume the instance is what we want instead of the type"""
                self.model = AbstractModel.Abstracted(ensure_model_not_instance(model))
                self.helper:SqlModelHelper = SqlModelHelper(model) 
            else: self.model=None
            self.conn = psycopg2.connect(self._connection_string)

        except:
            if on_connect_error != 'ignore':
                logger.warning(traceback.format_exc())
                logger.warning(
                    "Could not connect - you will need to check your env and call pg._connect again"
                )
            
    def _create_db(self,name:str):
        """this util is to create a test database primarily"""
        
        self.conn.autocommit = True   
        try:
            cursor = self.conn.cursor()
            cursor.execute(f"""SELECT 1 FROM pg_database WHERE datname = '{name}';""")
            exists = cursor.fetchone()

            if not exists:
                logger.info(f"Creating database: {name}")
                cursor.execute(f"CREATE DATABASE {name};")
            else:
                logger.debug(f"Database {name} already exists.")

            cursor.close()
            self.conn.close()
        except Exception as ex:
            logger.debug(ex)
            raise
        finally:
            pass
        
            
    def __repr__(self):
        return f"PostgresService({self.model.get_model_full_name() if self.model else None}, {POSTGRES_SERVER=}, {POSTGRES_DB=})"
            
    def _reopen_connection(self):
        """util to retry opening closed connections in the service"""
        @retry(wait=wait_fixed(1), stop=stop_after_attempt(4), reraise=True)
        def open_connection_with_retry(conn_string):
            return psycopg2.connect(conn_string, connect_timeout=5) 
        try:
            if self.conn is None:
                self.conn = open_connection_with_retry(self._connection_string)
            self.conn.poll()
        except psycopg2.InterfaceError as error:
            self.conn = None #until we can open it, lets not trust it
            self.conn = open_connection_with_retry(self._connection_string)

    def _connect(self):
        self.conn = psycopg2.connect(self._connection_string)
        return self.conn

    def repository(self, model: BaseModel, **kwargs) -> "PostgresService":
        """a connection in the context of the abstract model for crud support"""
        return PostgresService(model=model, connection_string=self._connection_string, **kwargs)

    def get_entities(self, keys: str | typing.List[str]):
        """
        use the get_entities database function to lookup entities
        """
        if keys:
            if not isinstance(keys,list):
                keys = [keys]
                
        data = self.execute(f"""SELECT * FROM p8.get_entities(%s)""", data=(keys,))     if keys else None  
            
        if not data:
            return [{
                "status": "NO DATA",
                "message": f"There were no data when we fetched {keys=} Please use another method to answer the question or return to the user with a new suggested plan or summary of what you know so far. If you still have different functions to use please try those before completion." 
            }]      
        return  data[0]
            

        
    def search(self, question:str):
        """
        If the repository has been activated with a model we use the models search function
        Otherwise we use percolates generic plan and search.
        Either way, feel free to ask a detailed question and we will seek data.
        
        Args:
            question: detailed natural language question 
        """
        
        """in future we should pardo multiple questions"""
        if isinstance(question,list):
            question = '\n'.join(question)
        
        Q = f"""select * from p8.query_entity(%s,%s) """
        
        return self.execute(Q, data=(question,self.model.get_model_full_name() ))
        
    def get_model_database_schema(self):
        assert self.model is not None, "The model is empty - you should construct an instance of the postgres service as a repository(Model)"
        q = f"""SELECT 
            column_name AS field_name,
            data_type AS field_type
            FROM information_schema.columns
            WHERE table_name = %s AND table_schema = %s
        """
        return self.execute(q, data=(self.model.get_model_name(),self.model.get_model_namespace()))
    
    def model_registration_script(self, primary: bool= True, secondary:bool=True):
        """for bootstrapping we can generate the model registration script including data
        normally we need to create all the tables first and then we can add aux stuff
        """
        primary_scripts = {
            'register entity':self.helper.create_script(if_not_exists=True),  
        }
        secondary_scripts = { 
            'register_embeddings': self.helper.create_embedding_table_script(),
            'insert_field_data': SqlModelHelper(ModelField).get_data_load_statement(self.helper.get_model_field_models()),
            'insert_agent_data': SqlModelHelper(Agent).get_data_load_statement(self.helper.get_model_agent_record()),
            'register_entities' : self.helper.get_register_entities_query()
        }
        
        
        statement = ''
        def add_scripts(scripts,s):
            for k,v in scripts.items():
                if v:
                    s += f'\n-- {k} ({self.helper.model.get_model_full_name()})------' 
                    s += '\n-- ------------------\n' 
                    s += v
                    s += '\n-- ------------------\n' 
            return s
        if primary:
            statement = add_scripts(primary_scripts,statement)
        if secondary:
            statement = add_scripts(secondary_scripts,statement)
            
        return statement
            
    def register(
        self,
        plan: bool = False,
        register_entities: bool = True,
        allow_create_schema: bool = False
    ):
        """register the entity in percolate
        -- create the type's table and embeddings table
        -- add the fields for the model
        -- add the agent model entity
        -- register the entity which means adding some supporting views etc.
        """
        assert (
            self.model is not None
        ), "You need to specify a model in the constructor or via a repository to register models"
        script = self.helper.create_script()
        #logger.debug(script)
        if plan == True:
            logger.debug(f"Exiting as {plan=}")
            return script

        try:
            self.execute(script,verbose_errors=False)
            logger.debug(f"Created table {self.helper.model.get_model_table_name()}")   
        except DuplicateTable:
            logger.warning(f"The table already exists - will check for schema migration or ignore")
            current_fields = self.get_model_database_schema()
            script = self.helper.try_generate_migration_script(current_fields)
            if script:
                logger.warning(f"Migrating schema with {script}")
                self.execute(script)
          
        """added the embedding but check if there are certainly embedding columns"""
        if self.helper.table_has_embeddings:
            try:
                script = self.helper.create_embedding_table_script()
                self.execute(script,verbose_errors=False)
                logger.debug(f"Created embedding table - {self.helper.model.get_model_embedding_table_name()}")
            except DuplicateTable:
                logger.warning(f"The embedding-associated table already exists")


        if register_entities:
            logger.debug("Updating model fields")
            self.repository(ModelField).update_records(self.helper.get_model_field_models())
            
            logger.debug(f"Adding the model agent")
            self.repository(Agent).update_records(self.helper.get_model_agent_record())
        
            """the registration"""
            if 'name' in self.helper.model.model_fields:
                #TODO: we need some way to map a key field for the graph. at the moment a name property is at least implicitly required. We would need this or a business key attribute in the database
                self.execute("select * from p8.register_entities(%s)", data=(self.helper.model.get_model_full_name(),))
            
                logger.info(f"Entity registered")
        else:
            logger.info("Done - register entities was disabled")
        
    def execute(
        cls,
        query: str,
        data: tuple = None,
        as_upsert: bool = False,
        page_size: int = 100,
        verbose_errors:bool=True,
        timeout_seconds: int = None
    ):
        """run any sql query - this works only for selects and transactional updates without selects
        
        Args:
            query: 
            data: tuple of args
            as_upsert: hint to use the upsert mode
            page_size: for upsert batching
            verbose_errors:
            timeout_seconds: keep this off but for testing multi turn you can set it to something but keep in mind we would use background workers for this
        """
        
        if cls.conn is None:
            cls._reopen_connection()
        if not query:
            return
        
        if timeout_seconds and isinstance(timeout_seconds,int):
            query = f"set statement_timeout = '{timeout_seconds}s';{query}"
        try:
            """we can reopen the connection if needed"""
            try:
                c = cls.conn.cursor()
            except:
                cls._reopen_connection()
                c = cls.conn.cursor()
                
            """prepare the query"""
            if as_upsert:
                psycopg2.extras.execute_values(
                    c, query, data, template=None, page_size=page_size
                )
            else:
                c.execute(query, data)

            if c.description:
                result = c.fetchall()
                """if we have and updated and read we can commit and send,
                otherwise we commit outside this block"""
                cls.conn.commit()
                column_names = [desc[0] for desc in c.description or []]
                result = [dict(zip(column_names, r)) for r in result]
                return result
            """case of upsert no-query transactions"""
            cls.conn.commit()
        except Exception as pex:
            msg = f"Failing to execute query {query} for model {cls.model} - Postgres error: {pex}, {data}"
            if not verbose_errors:
                msg = f"Failing to execute query model {cls.model} - {verbose_errors=} - {pex}"
            logger.warning(msg)
            cls.conn.rollback()
            raise
        finally:
            cls.conn.close()
            cls.conn = None

    def select(self, fields: typing.List[str] = None, **kwargs):
        """
        select based on the model and use kwargs as quality or in-list template predicates
        """
        assert (
            self.model is not None
        ), "You need to specify a model in the constructor or via a repository to select models"

        data = None
        if kwargs:
            data = tuple(kwargs.values())
        return self.execute(self.helper.select_query(fields, **kwargs), data=data)

    def get_by_name(cls, name: str, as_model:bool = False):
        """select model by name"""
        data =  cls.select(name=name)
        if data and as_model and cls.model:
            return [cls.model(**d) for d in data]
        return data

    def get_by_id(cls, id: str, as_model:bool = False):
        """select model by id"""
        data =  cls.select(id=id)
        if data and as_model and cls.model:
            return [cls.model(**d) for d in data]
        return data
    
    def select_to_model(self, fields: typing.List[str] = None):
        """
        like select except we construct the model objects
        """
        return [self.model.model_parse(d) for d in self.select(fields)]


    def execute_upsert(cls, query: str, data: tuple = None, page_size: int = 100):
        """run an upsert sql query"""
        return cls.execute(query, data=data, page_size=page_size, as_upsert=True)

    def update_records(
        self, records: typing.List[BaseModel], batch_size: int = 1000, index_entities: bool = False
    ):
        """records are updated using typed object relational mapping."""

        if records and not isinstance(records, list):
            records = [records]

        if self.model is None:
            """we encourage explicitly construct repository but we will infer"""
            return self.repository(records[0]).update_records(
                records=records, batch_size=batch_size
            )

        if len(records) > batch_size:
            logger.info(f"Saving  {len(records)} records in batches of {batch_size}")
            for batch in batch_collection(records, batch_size=batch_size):
                sample = self.update_records(batch, batch_size=batch_size,index_entities=index_entities)
            return sample

        data = [
            tuple(self.helper.serialize_for_db(r).values())
            for i, r in enumerate(records)
        ]

        if records:
            query = self.helper.upsert_query(batch_size=len(records))
            try:
                result = self.execute_upsert(query=query, data=data)
            except:
                logger.warning(f"Failing to run {query}")
                raise

            if index_entities:
                self.index_entities()
            return result
        else:
            logger.warning(f"Nothing to do - records is empty {records}")

    def index_entity_by_name(self, entity_name:str, id:uuid.UUID=None):
        """
        index entities - a session id can be passed in for the audit callback
        this is very much WIP - it may be this moves into background workers in the database
        """
        
        assert self.model is not None, "The model is null - did you mean to create a repository with the model first?"
        
        r1, r2 = None,None
        errors = ""
        try:
            
            r1 = self.execute(f""" select * from p8.insert_entity_nodes(%s); """, data=(entity_name,))
        except Exception as ex:
            errors+= traceback.format_exc()
            logger.warning(f"Failed to compute nodes {traceback.format_exc()}")
        try:
            r2 = self.execute(f""" select * from p8.insert_entity_embeddings(%s); """, data=(entity_name,))
        except Exception as ex:
            errors+= traceback.format_exc()
            logger.warning(f"Failed to compute embeddings {ex}")
            
        metrics =  {
            'entities added': r1,
            'embeddings added': r2,
        }
        
        if not id:
            id = uuid.uuid1()
        
        if errors == '':
            self.repository(IndexAudit).update_records(IndexAudit(id=id,
                                                                  model_name='percolate',
                                                                  entity_full_name=entity_name, 
                                                                  metrics=metrics, 
                                                                  status="OK", 
                                                                  message='Index updated without errors'))
        else:
            self.repository(IndexAudit).update_records(IndexAudit(id=id,
                                                                  model_name='percolate',
                                                                  entity_full_name=entity_name,
                                                                  metrics=metrics, 
                                                                  status="ERROR", 
                                                                  message=errors))
        
        logger.info(metrics)
        return metrics
    
    def index_entities(self):
        """This is to allow push index but we typically use the DB background workers to do this for us
        """
        
        logger.info(f'indexing entity {self.model}')
        return self.index_entity_by_name(self.model.get_model_full_name())
        


import typing
from percolate.models.p8 import Function, PlanModel
from percolate.models import AbstractModel
import percolate as p8
from percolate.utils import logger
from percolate.models import inspection
from pydantic import BaseModel

class _RuntimeFunction(Function):
    """A wrapper for handling library functions"""
    fn: typing.Callable
    
    def __call__(self, **kwargs):
        """overrides the proxied base call"""
        return self.fn(**kwargs)
    
class FunctionManager:
    def __init__(cls):
        cls._functions= {}
        cls.repo = p8.repository(Function)
        
    def __getitem__(cls, key):
        """unsafely gets the function"""
        return cls._functions[key]
    
    def add_function(cls, function: typing.Callable | Function):
        """add a function to the stack of functions given to the llm
        
            Args: a callable function or percolate Function type
        """
        
        if not isinstance(function, Function):
            logger.debug(f"adding function: {function}")
            function = _RuntimeFunction.from_callable(function)
        if function.name not in cls._functions:
            cls._functions[function.name] = function
            logger.debug(f"added function {function.name}")
    
    def activate_agent_context(cls, agent_model: AbstractModel):
        """
        Add and abstract model to activate functions on the type
        Pydantic BaseModels can be abstracted with the AbstractModel.Abstract if necessary
        In practice we import any callable methods on the agent model as callables
        
        Args:
            agent_model: Any class type with @classmethods that can be passed to the LLM
        """
        required = list((agent_model.get_model_functions() or {}).keys())
        for f in cls.repo.get_by_name(required, as_model=True):
            cls.add_function(f)
        required = set(required) - set(cls.functions.keys())
        if len(required):
            logger.warning(f"We could not find the function {required}")  
        
        """percolate is designed for external agents but we can support inline functions"""
        for f in inspection.get_class_and_instance_methods(agent_model, inheriting_from=agent_model):
            cls.add_function(f)
 
    def add_functions_by_key(cls, function_keys : typing.List[str]|str):
        """Add function or functions by key(s) - the function keys are expected to existing in the registry
        
        Args:
            function_keys: a list of one or more function(keys) to activate
            
        """
        if function_keys:
            if not isinstance(function_keys,list):
                function_keys = [function_keys]
        """activate here"""   
        required = [f for f in function_keys if f not in cls.functions]     
        if required:
            for f in cls.repo.get_by_name(required, as_model=True):
                cls.add_function(f)
        required = set(required) - set(cls.functions.keys())
        if required:
            logger.warning(f"We could not find the function {required}")
        
    def plan(cls, questions: str | typing.List[str], use_cache: bool = False):
        """based on one or more questions, we will construct a plan.
        This uses the database function plan to search over functions.
        We can also use the cache to load the functions into memory and do the same thing
        
        Args:
            questions: a question or list of questions to use to construct a plan over agents and functions
            use_cache: (default=False) use the in memory cache rather than the database function to make the plan
        """
        
        """TODO
        in the database we need a Plan model that also can search agents and return a plan
        but in python we can just select the data into the planner agent and fetch the plan
        """
        
        return p8.Agent(PlanModel,allow_help=False).run(questions, data=cls.repo.select())
    
    @property
    def functions(cls):
        return cls._functions
        
    @property
    def function_names(cls):
        return list(cls._functions.keys())


from .services.PostgresService import PostgresService
from .models import AbstractModel
import typing
from pydantic import BaseModel
from .services.ModelRunner import ModelRunner
from .services import OpenApiService
from .models.p8.db_types import AskResponse
import json

def dump(*args,**kwargs):
    """TODO:"""
    pass

def describe_agent(agent: AbstractModel | str, include_native_tools:bool=False):
    """
    Provides a description of the agent model as it would be passed to an LLM
    """
    
    if isinstance(agent,str):
        prompt = PostgresService().execute(f""" select * from p8.generate_markdown_prompt(%s) """, data=(agent,))
        prompt = prompt[0]['generate_markdown_prompt'] if prompt else None
        functions = PostgresService().execute(f""" select * from p8.get_agent_tools(%s,NULL,%s) """, data=(agent,include_native_tools))
        if functions:
            functions = functions[0]['get_agent_tools']
            #form canonical format
            functions = [f['function'] for f in functions]
            #to dict
            functions = {f['name']: f['description'] for f in functions}
        else:
            functions = {}
    else:
        agent = AbstractModel.Abstracted(agent)
        prompt =  agent.get_model_description()
        functions = agent.get_model_functions()
 
    function_desc = ""
    for k,v in (functions or {}).items():
        function_desc+=f""" - **{k}**: {v}\n"""

    prompt += f"""
    
## Functions
{function_desc}
    """
    
    return prompt
    

def get_entities(keys: str | typing.List)->typing.List[dict]:
    """
    get entities from their keys in the database
    
    **Args:
        keys: one or more keys 
    """

    data =  PostgresService().get_entities(keys)
 
    return data

def repository(model:AbstractModel|BaseModel):
    """gets a repository for the model. 
    This provides postgres services in the context of the type
    
    Args:
        model: a Pydantic base model or AbstractModel
    """
    return PostgresService(model)

def Agent(model:AbstractModel|BaseModel, **kwargs)->ModelRunner:
    """get the model runner in the context of the agent for running reasoning chains"""
    return ModelRunner(model,**kwargs)

def resume(session: AskResponse|str) ->AskResponse:
    """
    pass in a session id or ask response object to resume the session
    Resume session continues any non completed session
    """
    
    if isinstance(session,AskResponse):
        session = session.session_id
    
    response =  PostgresService().execute(f""" select * from p8.resume_session(%s); """, data=(session,) )    
    if response:
        print(response)
        return AskResponse(**response[0])
    else:
        raise Exception("Percolate gave no response")
    
    
def run(question: str, agent: str=None, limit_turns:int=2, **kwargs):
    """optional entry point to run an agent in the database by name
    The limit_turns controls how many turns are taken in the database e.g. call a tool and then ask  the agent to interpret 
    Args:
        question (str): any question for your agent
        agent: qualified agent name. Default schema is public and can be omitted - defaults to p8.PercolateAgent
        limit_turns: limit turns 2 allows for a single too call and interpretation for example
    """
    if not agent:
        agent = f"p8.PercolateAgent"
    elif '.' not in agent:
        agent = f"public.{agent}"
        
    response =  PostgresService().execute(f""" select * from percolate_with_agent(%s, %s); """, data=(question, agent) )    
    if response:
        print(response)
        return AskResponse(**response[0])
    else:
        raise Exception("Percolate gave no response")
    
def get_language_model_settings():
    """iterates through language models configured in the database.
    this is a convenience as you can also select * from p8."LanguageModelApi"
    """
    
    return PostgresService().execute('select * from p8."LanguageModelApi"')


def get_proxy(proxy_uri:str):
    """A proxy is a service that can call an external function such as an API or Database.
    We can theoretically proxy library functions but in python they should be added to the function manager as callables instead
    
    Args:
        proxy_uri: an openapi rest api or a native schema name for the database - currently the `p8` schema is assumed
    """
    if 'http' in proxy_uri or 'https' in proxy_uri:
        return OpenApiService(proxy_uri)
    if 'p8.' in proxy_uri:
        return PostgresService()
    
    raise NotImplemented("""We will add a default library proxy for the functions in the library 
                         but typically the should just be added at run time _as_ callables since 
                         we can recover Functions from callables""")
    
def get_planner()->typing.Callable:
    """retrieves a wrapper to the planner agent which takes a question for planning
    
    """
    from percolate.models.p8 import Function,PlanModel
    from functools import partial
    a = Agent(PlanModel,allow_help=False, init_data =repository(Function).select())
    return a


# routers/drafts.py
from fastapi import APIRouter, HTTPException, BackgroundTasks
from fastapi import   Depends, Response 
import json
router = APIRouter()
from percolate.api.auth import get_current_token
from pydantic import BaseModel, Field
import typing
import uuid
from percolate.services import PostgresService
from percolate.models.p8 import IndexAudit
from percolate.utils import logger
import traceback
from percolate.utils.studio import Project, apply_project

@router.post("/env/sync")
async def sync_env(user: dict = Depends(get_current_token)):
    """sync env adds whatever keys you have in your environment your database instance
    This is used on database setup or if keys are missing in database sessions
    """
    return Response(content=json.dumps({'status':'ok'}))


class AddApiRequest(BaseModel):
    uri: str = Field(description="Add the uri to the openapi.json for the API you want to add")
    token: typing.Optional[str] = Field(description="Add an optional bearer token or API key for API access")
    verbs: typing.Optional[str] = Field(description="A comma-separated list of verbs e.g. get,post to filter endpoints by when adding endpoints")
    endpoint_filter: typing.Optional[typing.List[str]] = Field(description="A list of endpoints to filter by when adding endpoints")
    
@router.post("/add/api")
async def add_api( request:AddApiRequest,  user: dict = Depends(get_current_token)):
    """add apis to Percolate
    """
    return Response(content=json.dumps({'status':'ok'}))

class AddAgentRequest(BaseModel):
    name: str = Field(description="A unique entity name, fully qualified by namespace or use 'public' as default" )
    functions: dict = Field(description="A mapping of function names in Percolate with a description of how the function is useful to you")
    spec: dict = Field(description="The Json spec of your agents structured response e.g. from a Pydantic model")
    description: str = Field(description="Your agent description - acts as a system prompt")
    
    
@router.post("/add/agent")
async def add_api( request:AddAgentRequest,  user: dict = Depends(get_current_token)):
    """add agents to Percolate. Agents require a Json Schema for any structured response you want to use, a system prompt and a dict/mapping of external registered functions.
    Functions can be registered via the add APIs endpoint.
    """
    return Response(content=json.dumps({'status':'ok'}))

@router.post("/add/project")
async def add_project( project: Project,  user: dict = Depends(get_current_token)):
    """Post the project yaml/json file to apply the settings. This can be used to add apis, agents and models. 
    
    - If you have set environment keys in your API we will sync these to your database if the `sync-env` flag is set in the project options
    - If you want to index the Percolation documentation set the flag `index-docs`
    """
    results = apply_project(project)
    return Response(content=json.dumps(results))


@router.get("/slow-endpoint")
async def slow_response():
    import time
    time.sleep(10)  # Simulate a delay
    return {"message": "This response was delayed by 10 seconds"}



class IndexRequest(BaseModel):
    """a request to update the indexes for entities by full name"""
    entity_full_name: str = Field(description="The full entity name - optionally omit for public namespace")

 
@router.post("/index/", response_model=IndexAudit)
async def index_entity(request: IndexRequest, background_tasks: BackgroundTasks, user: dict = Depends(get_current_token))->IndexAudit:
    """index entity and get an audit log id to check status
    the index is created as a background tasks and we respond with an id ref that can be used in the get/
    """
    id=uuid.uuid1()
    s = PostgresService(IndexAudit)
    try:
        
        record = IndexAudit(id=id, model_name='percolate', entity_full_name=request.entity_full_name, metrics={}, status="REQUESTED", message="Indexed requested")
        s.update_records(record)
        """todo create an audit record pending and use that in the api response"""
        background_tasks.add_task(s.index_entity_by_name, request.entity_full_name, id=id)
        return record
    except Exception as e:
        """handle api errors"""
        logger.warning(f"/admin/index {traceback.format_exc()}")
        record = IndexAudit(id=id,model_name='percolate',entity_full_name=request.entity_full_name, metrics={}, status="ERROR", message=str(e))
        """log the error"""
        s.update_records(record)
        raise HTTPException(status_code=500, detail="Failed to manage the index")
    
@router.get("/index/{id}", response_model=IndexAudit)
async def get_index(id: uuid.UUID) -> IndexAudit:
    """
    request the status of the index by id
    """
    #todo - proper error handling
    records =  PostgresService.get_by_id(id)
    if records:
        return records[0]
    """TODO error not found"""
    return {}


from pydantic import BaseModel
from percolate.models import AbstractModel
import json
import typing
class MessageStack:
    def __init__(self, question: str, system_prompt:str=None, data: typing.List[dict] = None):
        
        self.question = question
        self.system_prompt = system_prompt
        self.data = data or []

    def add(self, data: dict|typing.List[dict], **kwargs):
        """add messages to the stack such as function responses typically"""
        if not isinstance(data,list):
            data = [data]
        
        """if its a pydantic object thats fine too"""
        data = [d if not hasattr(d, 'model_dump') else d.model_dump() for d in data]
        
        self.data += data
    
    @classmethod
    def build_message_stack(cls, abstracted_model: AbstractModel, question:str,data: typing.List[dict] = None, **kwargs) ->"MessageStack":
        """
        we build a message stack from the model prompt and question
        
        Args:
            abstracted_model: provides at least a description for system prompt - we fall back to the doc string of any object
            question: the user question
            data: any initial data to load
        """
        if data:
            """need to think about the best way to add this"""
            data = [{
                "role": "user",
                "content": json.dumps(data,default=str)                
            }]
        generalized_prompt_preamble = kwargs.get('system_prompt_preamble')
        prompt = f"{generalized_prompt_preamble}\n{abstracted_model.get_model_description()}"
        return MessageStack(question=question, system_prompt=prompt, data = data)
        


import typing
import traceback
import percolate as p8
from pydantic import BaseModel
from percolate.utils import logger
from percolate.models.p8 import Function 
from .FunctionManager import FunctionManager
from percolate.models import AbstractModel, MessageStack
from percolate.services.llm import CallingContext, FunctionCall, LanguageModel, MessageStackFormatter

GENERIC_P8_PROMPT = """\n# General Advice.
Use whatever functions are available to you and use world knowledge only if prompted 
or if there is not other way 
or the user is obviously asking about real world information that is not covered by functions.
Observe what functions you have to use and check the message history to avoid calling the same functions with the same parameters repeatedly.
If you find a function name in your search, you can activate it by naming using one of your functions. You should do so without asking the user.
When generating large output observe the notification that needs to be called first, call it and continue.
            """
            
class ModelRunner:
    """The model runner manages chaining agents and reasoning steps together in the execution loop
    We had functions implementing the p8 model runner interface which is also in db native functions
    """
    
    @property
    def name(self):
        return self.agent_model.get_model_full_name()
    def __init__(self, model: BaseModel = None, allow_help: bool = True, **kwargs):
        """
        A model is passed in or the default is used.
        This supplies the agent context such as prompt and functions.
        If the model has no functions simple Q&A can still be exchanged with LLMs.
        More generally the model can provide a structured response format.
        see TODO: for guidance.
        """
        self._init_data = kwargs.get('init_data')
        """the agent model is any Pydantic Base model or Abstract model that implements the agent interface"""
        self.agent_model:AbstractModel = AbstractModel.Abstracted(model)
        """a function manager allows for activating functions at runtime and searching and planning over functions"""
        self._function_manager = FunctionManager()
        """the help option links in to the function manager planner"""
        self._allow_help = allow_help
        """the repository provides the Percolate database instance"""
        self.repo = p8.repository(self.agent_model)
        """initialize activates the agent model e.g. functions and prompt for use"""
        self.initialize()
        """the messages stack is the most important control element for llm agent sessions"""
        self.messages = MessageStack(None)
        logger.info(f"******Constructed agent {self.name}******")
        
    def __repr__(self):
        return f"Runner({self.name})"

    def initialize(self):
        """Register the functions and other metadata from the agent model"""
        self._context = None
        if self._allow_help:
            self._function_manager.add_function(self.help)  # :)
        """register the model's functions which can include function search"""
        self._function_manager.activate_agent_context(self.agent_model)
        """the basic bootstrapping means asking for help, entities(types) or functions"""
        self._function_manager.add_function(self.get_entities)
        self._function_manager.add_function(self.search)
        self._function_manager.add_function(self.activate_functions_by_name)
        self._function_manager.add_function(self.announce_generate_large_output)
        """more complex things will happen from here when we traverse what comes back"""

    def announce_generate_large_output(self, estimated_length:int=None):
        """When you are about to generate a lot of output, please call this function with a rough estimate of the size of the content.
        You do not need to do this when you are responding with simple structured responses which are typically small or with simple answers.
        However when generating lots of text we would like to request via streaming or async so we want to know before generating a lot of text.
        We use to distinguish internal content gathering nodes from final response generation for users.
        """
        logger.warning(f"The model is requesting to output: {estimated_length=}")
        return {'message': 'acknowledged', 'output_size_estimate': estimated_length}
        
        
    def search(self, questions: typing.List[str]):
        """Run a general search on the model that is being used in the current context as per the system prompt
        If you want to add multiple questions supply a list of strings as an array.
        Args:
            questions: ask one or more questions to search the data store
        """

        return self.repo.search(questions)

    def activate_functions_by_name(self, function_names: typing.List[str], **kwargs):
        """Provide a list of function names to load.
        The names should be fully qualified object_id.function_name
        """

        logger.debug(f"activating function {function_names}")
        _ = self._function_manager.add_functions_by_key(function_names)
        """todo check status"""
        return {
            "status": f"Re: the functions {function_names}, now ready for use. please go ahead and invoke."
        }

    def get_entities(self, keys: typing.Optional[str]):
        """Lookup entity by one or more keys. For example if you encounter entity names or keys in question, data etc you can use
        the entity search to learn more about them
        Args:
            keys: one or more names to use to lookup the entity or entities
        """
        logger.debug(f"get_entities/{keys=}")

        """the function manager can load context and we can also adorn entities with extra metadata"""
        entities = p8.get_entities(keys )

        return entities

    def help(self, questions: str | typing.List[str], context: str = None):
        """If you are stuck ask for help with very detailed questions to help the planner find resources for you.
        If you have a hint about what the source or tool to use hint that in each question that you ask.
        For example, you can either just ask a question or you can ask "according to resource X" and then ask the question. This is important context.

        Args:
            questions (str): provide detailed questions to guide the planner. you should augment the users question with the additional context that you have e.g. a known source
            context: any added context e.g. what tool, function, source the user or system suggests may know the answer
        """

        try:
            if context:
                questions = f"Using information from {context}, {questions}"

            """for now strict planning is off"""
            plan = self._function_manager.plan(questions)
        except Exception as ex:
            logger.warning(f"Failed to call help {ex}")
            return {"message": "planning pending - i suggest you use world knowledge"}

        """describe the plan context e.g. its a plan but you need to request the functions and do the thing -> update message stack"""

        return plan

    def invoke(self, function_call: FunctionCall):
        """Invoke function(s) and parse results into messages

        Args:
            function_call (FunctionCall): the payload send from an LLM to call a function
        """
        logger.debug(f"({self.name}){function_call=}")
        f = self._function_manager[function_call.name]
        if not f:
            message = f"attempting to load function {function_call.name} which is not activated - please activate it"
            data = MessageStackFormatter.format_function_response_error(
                function_call, ValueError(message), self._context
            )
        else:
            try:
                """try call the function - assumes its some sort of json thing that comes back"""
                data = f(**function_call.arguments) or {}
                data = MessageStackFormatter.format_function_response_data(
                    function_call, data, self._context
                )
                """if there is an error, how you format the message matters - some generic ones are added
                its important to make sure the format coincides with the language model being used in context
                """
            except TypeError as tex:  # type errors are usually the agents fault
                logger.warning(f"Error calling function {tex}")
                data = MessageStackFormatter.format_function_response_type_error(
                    function_call, tex, self._context
                )
            except Exception as ex:  # general errors are usually our fault
                logger.warning(f"Error calling function {traceback.format_exc()}")
                data = MessageStackFormatter.format_function_response_error(
                    function_call, ex, self._context
                )

        # print(data) # maybe trace here
        """update messages with data if we can or add error messages to notify the language model"""
        self.messages.add(data)

    @property
    def functions(self) -> typing.Dict[str, Function]:
        """Provide access to the function manager's functions"""
        return self._function_manager.functions
    
    @property
    def function_descriptions(self) -> typing.List[dict]:
        """Provide access to the function manager's function specs to send to language models"""
 
        return [f.function_spec for _,f  in self._function_manager.functions.items()]

    def __call__(
        self, question: str, context: CallingContext = None, limit: int = None,data: typing.List[dict] = None, language_model:str=None
    ):
        """
        Ask a question to kick of the agent loop
        """
        return self.run(question, context, data=data, limit=limit,language_model=language_model)


    def run(self, question: str, context: CallingContext = None, limit: int = None, data: typing.List[dict] = None, language_model:str=None):
        """Ask a question to kick of the agent loop
        
        Args:
            question: a user question
            context: a context object with information about users, session and usage
            limit: given that we iterate in the executor, we need to set a max length
            data: we can initialise the data payload as though there are function/data load results
        """
        
        """sometimes you may want to initialize your agent with a bunch of data"""
        data = data or self._init_data
        """setup all the bits before running the loop"""
        self._context = context or CallingContext.with_model(language_model)
        """a generic wrapper around the REST interfaces of any LLM client"""
        lm_client = LanguageModel.from_context(self._context)

        """messages are system prompt etc. agent model's can override how message stack is constructed
           we add p8 preamble in percolate - in future this could be disabled per model
        """
        self.messages = self.agent_model.build_message_stack(question=question, functions=self.functions.keys(), data=data,
                                                             system_prompt_preamble=GENERIC_P8_PROMPT)

        """run the agent loop to completion"""
        for _ in range(limit or self._context.max_iterations):
            response = None
            """the language model may stream into a callback in the calling context"""
            response = lm_client(
                messages=self.messages,
                context=self._context,
                functions=self.function_descriptions,
            )
            if function_calls := response.tool_calls:
                """models need us to add the tool call to the stack - this is not the case for openai function call but for consistency over models we must"""
                self.messages.add(response.verbatim)
                """call one or more functions and update messages - functions can be updated inside this context"""
                for func_call in function_calls: #its assumed to be only one for now but we could par do in future
                    self.invoke(FunctionCall(**func_call))
                continue
            if response is not None:
                # marks the fact that we have unfinished business
                break

        """fire telemetry"""
        p8.dump(question, response, self._context)

        return response.content




import os
from pathlib import Path
from importlib import import_module

def get_repo_root():
    """the root directory of the project by convention"""
    path = os.environ.get("P8_HOME")
    if not path:
        p8 = import_module("percolate")
        path = str(Path(p8.__file__).resolve().parent.parent.parent.parent.parent)
        return path
    return path
    
P8_HOME = os.environ.get('P8_HOME', get_repo_root())
STUDIO_HOME = f"{P8_HOME}/studio"
P8_SCHEMA = 'p8'
P8_EMBEDDINGS_SCHEMA = 'p8_embeddings'
#
POSTGRES_DB = "app"
POSTGRES_SERVER = os.environ.get('P8_PG_HOST', 'localhost')
POSTGRES_PORT = os.environ.get('P8_PG_PORT', 5438)
POSTGRES_PASSWORD = os.environ.get('P8_PG_PASSWORD', 'postgres') 
POSTGRES_USER = os.environ.get('P8_PG_USER', 'postgres') 
POSTGRES_CONNECTION_STRING = f"postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{POSTGRES_SERVER}:{POSTGRES_PORT}/{POSTGRES_DB}"
TESTDB_CONNECTION_STRING =  f"postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{POSTGRES_SERVER}:{POSTGRES_PORT}/test"
DEFAULT_CONNECTION_TIMEOUT = 30
#
MINIO_SECRET = os.environ.get('MINIO_SECRET', 'percolate')
MINIO_SERVER = os.environ.get('MINIO_SERVER', 'localhost:9000')
MINIO_P8_BUCKET = 'percolate'
#

GPT_MINI = "gpt-4o-mini"
DEFAULT_MODEL =   "gpt-4o-2024-08-06"

def load_db_key(key = "P8_API_KEY"):
    """valid database login requests the key for API access"""
    from percolate.services import PostgresService
    from percolate.utils import make_uuid
    pg = PostgresService()
    data = pg.execute(f'SELECT value from p8."Settings" where id = %s limit 1', (str(make_uuid({"key": key})),))
    if data:
        return data[0]['value']
    
    
def sync_model_keys(connection_string:str=None) -> dict:
    """look for any keys required and returns which there are and which are loaded in env"""
    from percolate.services import PostgresService
    pg = PostgresService(connection_string=connection_string)
    rows = pg.execute(f"""select distinct token_env_key from p8."LanguageModelApi" """)
    
    d = {}
    for row in rows:
        k = row['token_env_key'] 
        if token:= os.environ.get(k):
            d[k] = True
            pg.execute(f"""update p8."LanguageModelApi" set token=%s where token_env_key = %s""", data=(token,k))
        else:
            d[k] = False
    return d
        